import numpy as np
import logging
import itertools
from typing import Dict, Optional, List
from flaml.tune import Categorical, Float, PolynomialExpansionSet
from flaml.tune import Trial
from flaml.onlineml import VowpalWabbitTrial
from flaml.searcher import CFO

logger = logging.getLogger(__name__)


class BaseSearcher:
    """Implementation of the BaseSearcher

    Methods:
        set_search_properties(metric, mode, config)
        next_trial()
        on_trial_result(trial_id, result)
        on_trial_complete()
    """

    def __init__(self,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 ):
        pass

    def set_search_properties(self, metric: Optional[str] = None, mode: Optional[str] = None,
                              config: Optional[Dict] = None):
        if metric:
            self._metric = metric
        if mode:
            assert mode in ["min", "max"], "`mode` must be 'min' or 'max'."
            self._mode = mode

    def next_trial(self):
        NotImplementedError

    def on_trial_result(self, trial_id: str, result: Dict):
        pass

    def on_trial_complete(self, trial):
        pass


class ChampionFrontierSearcher(BaseSearcher):
    """The ChampionFrontierSearcher class

    Methods:
        (metric, mode, config)
            Generate a list of new challengers, and add them to the _challenger_list
        next_trial()
            Pop a trial from the _challenger_list
        on_trial_result(trial_id, result)
            Doing nothing
        on_trial_complete()
            Doing nothing

    NOTE:
        This class serves the role of ConfigOralce.
        Every time we create an online trial, we generate a searcher_trial_id.
        At the same time, we also record the trial_id of the VW trial.
        Note that the trial_id is a unique signature of the configuraiton.
        So if two VWTrials are associated with the same config, they will have the same trial_id
        (although not the same searcher_trial_id).
        searcher_trial_id will be used in suggest()
    """
    # ****the following constants are used when generating new challengers in
    # the _query_config_oracle function
    # how many item to add when doing the expansion
    # (i.e. how many interaction items to add at each time)
    POLY_EXPANSION_ADDITION_NUM = 1
    # the order of polynomial expansions to add based on the given seed interactions
    EXPANSION_ORDER = 2
    # the number of new challengers with new numerical hyperparamter configs
    NUMERICAL_NUM = 2

    # In order to use CFO, a loss name and loss values of configs are need
    # since CFO in fact only requires relative loss order of two configs to perform
    # the update, a pseudo loss can be used as long as the relative performance orders
    # of different configs are perserved. We set the loss of the init config to be
    # a large value (CFO_SEARCHER_LARGE_LOSS), and set the loss of the better config as
    # 0.95 of the previous best config's loss.
    # NOTE: this setting depends on the assumption that  (and thus
    # _query_config_oracle) is only triggered when a better champion is found.
    CFO_SEARCHER_METRIC_NAME = 'pseudo_loss'
    CFO_SEARCHER_LARGE_LOSS = 1e6

    # the random seed used in generating numerical hyperparamter configs (when CFO is not used)
    NUM_RANDOM_SEED = 111

    CHAMPION_TRIAL_NAME = 'champion_trial'
    TRIAL_CLASS = VowpalWabbitTrial

    def __init__(self,
                 init_config: Dict,
                 space: Optional[Dict] = None,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 random_seed: Optional[int] = 2345,
                 online_trial_args: Optional[Dict] = {},
                 nonpoly_searcher_name: Optional[str] = 'CFO'
                 ):
        '''Constructor

        Args:
            init_config: dict
            space: dict
            metric: str
            mode: str
            random_seed: int
            online_trial_args: dict
            nonpoly_searcher_name: A string to specify the search algorithm
                for nonpoly hyperparameters
        '''
        self._init_config = init_config
        self._space = space
        self._seed = random_seed
        self._online_trial_args = online_trial_args
        self._nonpoly_searcher_name = nonpoly_searcher_name

        self._random_state = np.random.RandomState(self._seed)
        self._searcher_for_nonpoly_hp = {}
        self._space_of_nonpoly_hp = {}
        # dicts to remember the mapping between searcher_trial_id and trial_id
        self._searcher_trialid_to_trialid = {}  # key: searcher_trial_id, value: trial_id
        self._trialid_to_searcher_trial_id = {}  # value: trial_id, key: searcher_trial_id
        self._challenger_list = []
        # initialize the search in set_search_properties
        self.set_search_properties(config={self.CHAMPION_TRIAL_NAME: None}, init_call=True)
        logger.debug('using random seed %s in config oracle', self._seed)

    def set_search_properties(self, metric: Optional[str] = None,
                              mode: Optional[str] = None,
                              config: Optional[Dict] = {},
                              init_call: Optional[bool] = False):
        """Construct search space with given config, and setup the search
        """
        super().set_search_properties(metric, mode, config)
        # *********Use ConfigOralce (i.e, self._generate_new_space to generate list of new challengers)
        logger.info('champion trial %s', config)
        champion_trial = config.get(self.CHAMPION_TRIAL_NAME, None)
        if champion_trial is None:
            champion_trial = self._create_trial_from_config(self._init_config)
        # generate a new list of challenger trials
        new_challenger_list = self._query_config_oracle(champion_trial.config,
                                                        champion_trial.trial_id,
                                                        self._trialid_to_searcher_trial_id[champion_trial.trial_id])
        # add the newly generated challengers to existing challengers
        # there can be duplicates and we check duplicates when calling next_trial()
        self._challenger_list = self._challenger_list + new_challenger_list
        # add the champion as part of the new_challenger_list when called initially
        if init_call:
            self._challenger_list.append(champion_trial)
        logger.critical('Created challengers from champion %s', champion_trial.trial_id)
        logger.critical('New challenger size %s, %s', len(self._challenger_list),
                        [t.trial_id for t in self._challenger_list])

    def next_trial(self):
        """Return a trial from the _challenger_list
        """
        next_trial = None
        if self._challenger_list:
            next_trial = self._challenger_list.pop()
        return next_trial

    def _create_trial_from_config(self, config, searcher_trial_id=None):
        if searcher_trial_id is None:
            searcher_trial_id = Trial.generate_id()
        trial = self.TRIAL_CLASS(config, **self._online_trial_args)
        self._searcher_trialid_to_trialid[searcher_trial_id] = trial.trial_id
        # only update the dict when the trial_id does not exist
        if trial.trial_id not in self._trialid_to_searcher_trial_id:
            self._trialid_to_searcher_trial_id[trial.trial_id] = searcher_trial_id
        return trial

    def _query_config_oracle(self, seed_config, seed_config_trial_id,
                             seed_config_searcher_trial_id=None) -> List[Trial]:
        """Give the seed config, generate a list of new configs (which are supposed to include
        at least one config that has better performance than the input seed_config)
        """
        # group the hyperparameters according to whether the configs of them are independent
        # with the other hyperparameters
        hyperparameter_config_groups = []
        searcher_trial_ids_groups = []
        nonpoly_config = {}
        for k, v in seed_config.items():
            config_domain = self._space[k]
            if isinstance(config_domain, PolynomialExpansionSet):
                # get candidate configs for hyperparameters of the PolynomialExpansionSet type
                partial_new_configs = self._generate_independent_hp_configs(k, v, config_domain)
                if partial_new_configs:
                    hyperparameter_config_groups.append(partial_new_configs)
                    # does not have searcher_trial_ids
                    searcher_trial_ids_groups.append([])
            elif isinstance(config_domain, Float) or isinstance(config_domain, Categorical):
                # otherwise we need to deal with them in group
                nonpoly_config[k] = v
                if k not in self._space_of_nonpoly_hp:
                    self._space_of_nonpoly_hp[k] = self._space[k]

        # -----------generate partial new configs for non-PolynomialExpansionSet hyperparameters
        if nonpoly_config:
            new_searcher_trial_ids = []
            partial_new_nonpoly_configs = []
            if 'CFO' in self._nonpoly_searcher_name:
                if seed_config_trial_id not in self._searcher_for_nonpoly_hp:
                    self._searcher_for_nonpoly_hp[seed_config_trial_id] = CFO(space=self._space_of_nonpoly_hp,
                                                                              points_to_evaluate=[nonpoly_config],
                                                                              metric=self.CFO_SEARCHER_METRIC_NAME,
                                                                              )
                    # initialize the search in set_search_properties
                    self._searcher_for_nonpoly_hp[seed_config_trial_id].set_search_properties(
                        config={'metric_target': self.CFO_SEARCHER_LARGE_LOSS})
                    # We need to call this for once, such that the seed config in points_to_evaluate will be called
                    # to be tried
                    self._searcher_for_nonpoly_hp[seed_config_trial_id].suggest(seed_config_searcher_trial_id)
                # assuming minimization
                if self._searcher_for_nonpoly_hp[seed_config_trial_id].metric_target is None:
                    pseudo_loss = self.CFO_SEARCHER_LARGE_LOSS
                else:
                    pseudo_loss = self._searcher_for_nonpoly_hp[seed_config_trial_id].metric_target * 0.95
                pseudo_result_to_report = {}
                for k, v in nonpoly_config.items():
                    pseudo_result_to_report['config/' + str(k)] = v
                pseudo_result_to_report[self.CFO_SEARCHER_METRIC_NAME] = pseudo_loss
                pseudo_result_to_report['time_total_s'] = 1
                self._searcher_for_nonpoly_hp[seed_config_trial_id].on_trial_complete(seed_config_searcher_trial_id,
                                                                                      result=pseudo_result_to_report)
                while len(partial_new_nonpoly_configs) < self.NUMERICAL_NUM:
                    # suggest multiple times
                    new_searcher_trial_id = Trial.generate_id()
                    new_searcher_trial_ids.append(new_searcher_trial_id)
                    suggestion = self._searcher_for_nonpoly_hp[seed_config_trial_id].suggest(new_searcher_trial_id)
                    if suggestion is not None:
                        partial_new_nonpoly_configs.append(suggestion)
                logger.info('partial_new_nonpoly_configs %s', partial_new_nonpoly_configs)
            else:
                raise NotImplementedError
            if partial_new_nonpoly_configs:
                hyperparameter_config_groups.append(partial_new_nonpoly_configs)
                searcher_trial_ids_groups.append(new_searcher_trial_ids)
        # ----------- coordinate generation of new challengers in the case of multiple groups
        new_trials = []
        for i in range(len(hyperparameter_config_groups)):
            logger.info('hyperparameter_config_groups[i] %s %s',
                        len(hyperparameter_config_groups[i]),
                        hyperparameter_config_groups[i])
            for j, new_partial_config in enumerate(hyperparameter_config_groups[i]):
                new_seed_config = seed_config.copy()
                new_seed_config.update(new_partial_config)
                # For some groups of the hyperparameters, we may have already generated the
                # searcher_trial_id. In that case, we only need to retrieve the searcher_trial_id
                # instead of generating it again. So we do not generate searcher_trial_id and
                # instead set the searcher_trial_id to be None. When creating a trial from a config,
                # a searcher_trial_id will be generated if None is provided.
                # TODO: An alternative option is to generate a searcher_trial_id for each partial config
                if searcher_trial_ids_groups[i]:
                    new_searcher_trial_id = searcher_trial_ids_groups[i][j]
                else:
                    new_searcher_trial_id = None
                new_trial = self._create_trial_from_config(new_seed_config, new_searcher_trial_id)
                new_trials.append(new_trial)
        logger.info('new_configs %s', [t.trial_id for t in new_trials])
        return new_trials

    def _generate_independent_hp_configs(self, hp_name, current_config_value, config_domain) -> List:
        if isinstance(config_domain, PolynomialExpansionSet):
            seed_interactions = list(current_config_value) + list(config_domain.init_monomials)
            logger.critical('Seed namespaces (singletons and interactions): %s', seed_interactions)
            logger.info('current_config_value %s %s', current_config_value, seed_interactions)
            configs = self._generate_poly_expansion_sets(seed_interactions,
                                                         self.EXPANSION_ORDER,
                                                         config_domain.allow_self_inter,
                                                         config_domain.highest_poly_order,
                                                         self.POLY_EXPANSION_ADDITION_NUM,
                                                         )
        else:
            raise NotImplementedError
        configs_w_key = [{hp_name: hp_config} for hp_config in configs]
        return configs_w_key

    def _generate_poly_expansion_sets(self, seed_interactions, order, allow_self_inter,
                                      highest_poly_order, interaction_num_to_add):
        champion_all_combinations = self._generate_all_comb(seed_interactions, order, allow_self_inter, highest_poly_order)
        space = sorted(list(itertools.combinations(
                       champion_all_combinations, interaction_num_to_add)))
        self._random_state.shuffle(space)
        candidate_configs = [set(seed_interactions) | set(item) for item in space]
        final_candidate_configs = []
        for c in candidate_configs:
            new_c = set([e for e in c if len(e) > 1])
            final_candidate_configs.append(new_c)
        return final_candidate_configs

    @staticmethod
    def _generate_all_comb(seed_interactions: list, seed_interaction_order: int,
                           allow_self_inter: Optional[bool] = False,
                           highest_poly_order: Optional[int] = None):
        """Generate new interactions by doing up to seed_interaction_order on the seed_interactions

        Args:
            seed_interactions (List[str]): the see config which is a list of interactions string
            (including the singletons)
            seed_interaction_order (int): the maxmum order of interactions to perform on the seed_config
            allow_self_inter (bool): whether self-interaction is allowed
                e.g. if set False, 'aab' will be considered as 'ab', i.e. duplicates in the interaction
                string are removed.
            highest_poly_order (int): the highest polynomial order allowed for the resulting interaction.
                e.g. if set 3, the interaction 'abcd' will be excluded.
        """

        def get_interactions(list1, list2):
            """Get combinatorial list of tuples
            """
            new_list = []
            for i in list1:
                for j in list2:
                    # each interaction is sorted. E.g. after sorting
                    # 'abc' 'cba' 'bca' are all 'abc'
                    # this is done to ensure we can use the config as the signature
                    # of the trial, i.e., trial id.
                    new_interaction = ''.join(sorted(i + j))
                    if new_interaction not in new_list:
                        new_list.append(new_interaction)
            return new_list

        def strip_self_inter(s):
            """Remove duplicates in an interaction string
            """
            if len(s) == len(set(s)):
                return s
            else:
                # return ''.join(sorted(set(s)))
                new_s = ''
                char_list = []
                for i in s:
                    if i not in char_list:
                        char_list.append(i)
                        new_s += i
                return new_s

        interactions = seed_interactions.copy()
        all_interactions = []
        while seed_interaction_order > 1:
            interactions = get_interactions(interactions, seed_interactions)
            seed_interaction_order -= 1
            all_interactions += interactions
        if not allow_self_inter:
            all_interactions_no_self_inter = []
            for s in all_interactions:
                s_no_inter = strip_self_inter(s)
                if len(s_no_inter) > 1 and s_no_inter not in all_interactions_no_self_inter:
                    all_interactions_no_self_inter.append(s_no_inter)
            all_interactions = all_interactions_no_self_inter
        if highest_poly_order is not None:
            all_interactions = [c for c in all_interactions if len(c) <= highest_poly_order]
        logger.info('all_combinations %s', all_interactions)
        return all_interactions



'''
Copyright 2020 The Ray Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This source file is adapted here because ray does not fully support Windows.

Copyright (c) Microsoft Corporation.
'''
from typing import Dict, Optional
import numpy as np
from .trial import Trial

import logging
logger = logging.getLogger(__name__)


def is_nan_or_inf(value):
    return np.isnan(value) or np.isinf(value)


class ExperimentAnalysis:
    """Analyze results from a Tune experiment.
    """

    @property
    def best_trial(self) -> Trial:
        """Get the best trial of the experiment
        The best trial is determined by comparing the last trial results
        using the `metric` and `mode` parameters passed to `tune.run()`.
        If you didn't pass these parameters, use
        `get_best_trial(metric, mode, scope)` instead.
        """
        if not self.default_metric or not self.default_mode:
            raise ValueError(
                "To fetch the `best_trial`, pass a `metric` and `mode` "
                "parameter to `tune.run()`. Alternatively, use the "
                "`get_best_trial(metric, mode)` method to set the metric "
                "and mode explicitly.")
        return self.get_best_trial(self.default_metric, self.default_mode)

    @property
    def best_config(self) -> Dict:
        """Get the config of the best trial of the experiment
        The best trial is determined by comparing the last trial results
        using the `metric` and `mode` parameters passed to `tune.run()`.
        If you didn't pass these parameters, use
        `get_best_config(metric, mode, scope)` instead.
        """
        if not self.default_metric or not self.default_mode:
            raise ValueError(
                "To fetch the `best_config`, pass a `metric` and `mode` "
                "parameter to `tune.run()`. Alternatively, use the "
                "`get_best_config(metric, mode)` method to set the metric "
                "and mode explicitly.")
        return self.get_best_config(self.default_metric, self.default_mode)

    def _validate_metric(self, metric: str) -> str:
        if not metric and not self.default_metric:
            raise ValueError(
                "No `metric` has been passed and  `default_metric` has "
                "not been set. Please specify the `metric` parameter.")
        return metric or self.default_metric

    def _validate_mode(self, mode: str) -> str:
        if not mode and not self.default_mode:
            raise ValueError(
                "No `mode` has been passed and  `default_mode` has "
                "not been set. Please specify the `mode` parameter.")
        if mode and mode not in ["min", "max"]:
            raise ValueError("If set, `mode` has to be one of [min, max]")
        return mode or self.default_mode

    def get_best_trial(self,
                       metric: Optional[str] = None,
                       mode: Optional[str] = None,
                       scope: str = "last",
                       filter_nan_and_inf: bool = True) -> Optional[Trial]:
        """Retrieve the best trial object.
        Compares all trials' scores on ``metric``.
        If ``metric`` is not specified, ``self.default_metric`` will be used.
        If `mode` is not specified, ``self.default_mode`` will be used.
        These values are usually initialized by passing the ``metric`` and
        ``mode`` parameters to ``tune.run()``.
        Args:
            metric (str): Key for trial info to order on. Defaults to
                ``self.default_metric``.
            mode (str): One of [min, max]. Defaults to ``self.default_mode``.
            scope (str): One of [all, last, avg, last-5-avg, last-10-avg].
                If `scope=last`, only look at each trial's final step for
                `metric`, and compare across trials based on `mode=[min,max]`.
                If `scope=avg`, consider the simple average over all steps
                for `metric` and compare across trials based on
                `mode=[min,max]`. If `scope=last-5-avg` or `scope=last-10-avg`,
                consider the simple average over the last 5 or 10 steps for
                `metric` and compare across trials based on `mode=[min,max]`.
                If `scope=all`, find each trial's min/max score for `metric`
                based on `mode`, and compare trials based on `mode=[min,max]`.
            filter_nan_and_inf (bool): If True (default), NaN or infinite
                values are disregarded and these trials are never selected as
                the best trial.
        """
        metric = self._validate_metric(metric)
        mode = self._validate_mode(mode)

        if scope not in ["all", "last", "avg", "last-5-avg", "last-10-avg"]:
            raise ValueError(
                "ExperimentAnalysis: attempting to get best trial for "
                "metric {} for scope {} not in [\"all\", \"last\", \"avg\", "
                "\"last-5-avg\", \"last-10-avg\"]. "
                "If you didn't pass a `metric` parameter to `tune.run()`, "
                "you have to pass one when fetching the best trial.".format(
                    metric, scope))
        best_trial = None
        best_metric_score = None
        for trial in self.trials:
            if metric not in trial.metric_analysis:
                continue

            if scope in ["last", "avg", "last-5-avg", "last-10-avg"]:
                metric_score = trial.metric_analysis[metric][scope]
            else:
                metric_score = trial.metric_analysis[metric][mode]

            if filter_nan_and_inf and is_nan_or_inf(metric_score):
                continue

            if best_metric_score is None:
                best_metric_score = metric_score
                best_trial = trial
                continue

            if (mode == "max") and (best_metric_score < metric_score):
                best_metric_score = metric_score
                best_trial = trial
            elif (mode == "min") and (best_metric_score > metric_score):
                best_metric_score = metric_score
                best_trial = trial

        if not best_trial:
            logger.warning(
                "Could not find best trial. Did you pass the correct `metric` "
                "parameter?")
        return best_trial

    def get_best_config(self,
                        metric: Optional[str] = None,
                        mode: Optional[str] = None,
                        scope: str = "last") -> Optional[Dict]:
        """Retrieve the best config corresponding to the trial.
        Compares all trials' scores on `metric`.
        If ``metric`` is not specified, ``self.default_metric`` will be used.
        If `mode` is not specified, ``self.default_mode`` will be used.
        These values are usually initialized by passing the ``metric`` and
        ``mode`` parameters to ``tune.run()``.
        Args:
            metric (str): Key for trial info to order on. Defaults to
                ``self.default_metric``.
            mode (str): One of [min, max]. Defaults to ``self.default_mode``.
            scope (str): One of [all, last, avg, last-5-avg, last-10-avg].
                If `scope=last`, only look at each trial's final step for
                `metric`, and compare across trials based on `mode=[min,max]`.
                If `scope=avg`, consider the simple average over all steps
                for `metric` and compare across trials based on
                `mode=[min,max]`. If `scope=last-5-avg` or `scope=last-10-avg`,
                consider the simple average over the last 5 or 10 steps for
                `metric` and compare across trials based on `mode=[min,max]`.
                If `scope=all`, find each trial's min/max score for `metric`
                based on `mode`, and compare trials based on `mode=[min,max]`.
        """
        best_trial = self.get_best_trial(metric, mode, scope)
        return best_trial.config if best_trial else None



from collections import OrderedDict
from .grid_searchspace_auto import AutoGridSearchSpace


def hpo_space_custom(model_type=None,
                     model_size_type=None,
                     dataset_name_list: list = None,
                     subdataset_name=None,
                     algo_mode=None,
                     **custom_hpo_args):
    """
    The 5 arguments here cannot be deleted, they need to be kept consistent with
    other functions in HPO_SEARCH_SPACE_MAPPING
    """
    assert "hpo_space" in custom_hpo_args
    custom_search_space = custom_hpo_args["hpo_space"]
    return custom_search_space


def bounded_gridunion(model_type=None,
                      model_size_type=None,
                      dataset_name_list: list = None,
                      subdataset_name=None,
                      algo_mode=None,
                      **custom_hpo_args):
    assert "bound" in custom_hpo_args
    gridunion_space = HPO_SEARCH_SPACE_MAPPING["uni"](model_type,
                                                      model_size_type,
                                                      dataset_name_list,
                                                      subdataset_name,
                                                      **custom_hpo_args)
    for each_key in custom_hpo_args["bound"].keys():
        if "u" in custom_hpo_args["bound"][each_key]:
            upper = custom_hpo_args["bound"][each_key]["u"]
        else:
            upper = 100000
        if "l" in custom_hpo_args["bound"][each_key]:
            lower = custom_hpo_args["bound"][each_key]["l"]
        else:
            lower = -100000
        original_space = sorted(gridunion_space[each_key])
        upper_id = len(original_space)
        for x in range(len(original_space)):
            if original_space[x] > upper:
                upper_id = x
                break
        lower_id = 0
        for x in range(len(original_space) - 1, -1, -1):
            if original_space[x] < lower:
                lower_id = x
                break
        gridunion_space[each_key] = original_space[lower_id:upper_id]
    return gridunion_space


def hpo_space_gridunion(model_type=None,
                        model_size_type=None,
                        dataset_name_list: list = None,
                        subdataset_name=None,
                        algo_mode=None,
                        **custom_hpo_args):
    output_config = {}
    for each_model_type in ["bert", "roberta", "electra"]:
        # if each_model_type == model_type: continue
        this_config = AutoGridSearchSpace.from_model_and_dataset_name(
            each_model_type, model_size_type, dataset_name_list, subdataset_name, "hpo")
        from ..utils import merge_dicts
        output_config = merge_dicts(output_config, this_config)
        default_values = {}
        """
        adding the default configuration from transformers/training_args.py into hpo space
        """
        from transformers import TrainingArguments
        training_args = TrainingArguments(output_dir=".")
        for each_hp in output_config.keys():
            try:
                default_values[each_hp] = [getattr(training_args, each_hp)]
            except AttributeError:
                print("training args do not contain {}, passed".format(each_hp))

        output_config = merge_dicts(output_config, default_values)

    return output_config


def hpo_space_gridunion_smoke_test(
        model_type=None,
        model_size_type=None,
        dataset_name_list: list = None,
        subdataset_name=None,
        algo_mode=None,
        **custom_hpo_args):
    return {
        "learning_rate": {"l": 1e-6, "u": 1e-3, "space": "log"},
        "num_train_epochs": [0.01],
        "per_device_train_batch_size": [2],
        "warmup_ratio": {"l": 0.0, "u": 0.3, "space": "linear"},
        "weight_decay": {"l": 0.0, "u": 0.3, "space": "linear"}
    }


def hpo_space_generic(model_type=None,
                      model_size_type=None,
                      dataset_name_list: list = None,
                      subdataset_name=None,
                      algo_mode=None,
                      **custom_hpo_args):
    output_config = {
        "learning_rate": {"l": 1e-6, "u": 1e-3, "space": "log"},
        "num_train_epochs": {"l": 1.0, "u": 10.0, "space": "log"},
        "per_device_train_batch_size": [4, 8, 16, 32, 48],
        "warmup_ratio": {"l": 0.0, "u": 0.3, "space": "linear"},
        "weight_decay": {"l": 0.0, "u": 0.3, "space": "linear"}
    }
    return output_config


def hpo_space_generic_grid(model_type=None,
                           model_size_type=None,
                           dataset_name_list: list = None,
                           subdataset_name=None,
                           algo_mode=None,
                           **custom_hpo_args):
    output_config = {
        "learning_rate": [1e-5, 2e-5, 3e-5, 4e-5, 5e-5, 1e-4, 1.5e-4],
        "num_train_epochs": [3, 10],
        "per_device_train_batch_size": [16, 32],
        "warmup_ratio": [0, 0.06, 0.1],
        "weight_decay": [0, 0.1]
    }
    return output_config


def hpo_space_small(model_type=None,
                    model_size_type=None,
                    dataset_name_list: list = None,
                    subdataset_name=None,
                    algo_mode=None,
                    **custom_hpo_args):
    config_json = AutoGridSearchSpace.from_model_and_dataset_name(
        model_type, model_size_type, dataset_name_list, subdataset_name, "hpo")
    output_config = {}

    for each_hp in config_json.keys():
        if each_hp == "learning_rate":
            if len(config_json[each_hp]) > 1:
                output_config[each_hp] = {"l": 3e-5, "u": 1.5e-4, "space": "log"}
            else:
                output_config[each_hp] = config_json[each_hp]
        elif each_hp == "num_train_epochs":
            output_config[each_hp] = {"l": 2.0, "u": 4.0, "space": "linear"}
        elif each_hp == "per_device_train_batch_size":
            output_config[each_hp] = [16, 32, 64]
        elif each_hp == "warmup_ratio":
            output_config[each_hp] = {"l": 0.0, "u": 0.2, "space": "linear"}
        elif each_hp == "weight_decay":
            output_config[each_hp] = {"l": 0.0, "u": 0.3, "space": "linear"}
        else:
            output_config[each_hp] = config_json[each_hp]

    return output_config


def hpo_space_grid(model_type=None,
                   model_size_type=None,
                   dataset_name_list: list = None,
                   subdataset_name=None,
                   algo_mode=None,
                   **custom_hpo_args):
    return AutoGridSearchSpace.from_model_and_dataset_name(model_type,
                                                           model_size_type,
                                                           dataset_name_list,
                                                           subdataset_name,
                                                           algo_mode
                                                           )


HPO_SEARCH_SPACE_MAPPING = OrderedDict(
    [
        ("grid", hpo_space_grid),
        ("uni", hpo_space_gridunion),
        ("gnr", hpo_space_generic),
        ("uni_test", hpo_space_gridunion_smoke_test),
        ("cus", hpo_space_custom),
        ("buni", bounded_gridunion),
    ]
)


class AutoHPOSearchSpace:
    """
    This is a class for getting the hpo search space based on the search space mode
    (a string variable) instantiated as one of the HPO search spaces of the library when
    created with the `~flaml.nlp.hpo.AutoHPOSearchSpace.from_model_and_dataset_name` method.

    This class cannot be instantiated directly using ``__init__()`` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoHPOSearchSpace is designed to be instantiated "
            "using the `AutoHPOSearchSpace.from_config_and_method_name(cls, hpo_searchspace_name,"
            "model_type,model_size_type,dataset_name,subdataset_name=None,**custom_hpo_args)` methods."
        )

    @classmethod
    def from_model_and_dataset_name(cls,
                                    hpo_searchspace_mode,
                                    model_type,
                                    model_size_type,
                                    dataset_name_list: list = None,
                                    subdataset_name=None,
                                    algo_mode=None,
                                    **custom_hpo_args):
        """
        Instantiate one of the classes for getting the hpo search space from the search space name, model type,
        model size type, dataset name and sub dataset name

        Args:

            hpo_searchspace_mode:
                A string variable which is the mode of the hpo search space, it must be chosen from the following options:
                    - uni: the union of BERT, RoBERTa and Electra's grid configs
                    - grid: the recommended grid config of the LM specified in jobconfig.pre
                    - gnr: the generic continuous search space
                    - uni_test: the search space for smoke test
                    - cus: user customized search space, specified in the "hpo_space" argument in AutoTransformers.fit
                    - buni: bounded grid union search space

            model_type:
                A string variable which is the type of the model, e.g., "electra"

            model_size_type:
                A string variable which is the type of the model size, e.g., "small"

            dataset_name:
                A string variable which is the dataset name, e.g., "glue"

            subdataset_name:
                A string variable which is the sub dataset name,e.g., "rte"

            custom_hpo_args:
                Any additional keyword argument to be used for the function for the HPO search space

        Example:
            >>> AutoHPOSearchSpace.from_model_and_dataset_name("uni", "electra", "small", ["glue"], "rte", "hpo")
        """

        if hpo_searchspace_mode in HPO_SEARCH_SPACE_MAPPING.keys():
            hpo_space = HPO_SEARCH_SPACE_MAPPING[hpo_searchspace_mode](
                model_type,
                model_size_type,
                dataset_name_list,
                subdataset_name,
                algo_mode,
                **custom_hpo_args)
            return hpo_space
        raise ValueError(
            "Unrecognized method {},{} for this kind of AutoHPOSearchSpace: {}.\n"
            "Method name should be one of {}.".format(
                hpo_searchspace_mode, dataset_name_list, cls.__name__,
                ", ".join(HPO_SEARCH_SPACE_MAPPING.keys())
            )
        )



import argparse
import json
import os
import pathlib
import re
from dataclasses import dataclass, field


def dataset_subdataset_name_format_check(val_str):
    regex = re.compile(r"^[^:]*:[^:]*$")
    if (val_str is not None) and (not regex.search(val_str)):
        raise argparse.ArgumentTypeError("dataset_subdataset_name must be in the format {data_name}:{subdata_name}")
    return val_str


def pretrained_model_size_format_check(val_str):
    regex = re.compile(r"^[^:]*:(small|base|large|xlarge)")
    if (val_str is not None) and (not regex.search(val_str)):
        raise argparse.ArgumentTypeError("pretrained_model_size must be in the format {model_name}:{model_size},"
                                         "where {model_name} is the name from huggingface.co/models, {model_size}"
                                         "is chosen from small, base, large, xlarge")
    return val_str


def load_dft_args():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('--server_name', type=str, help='server name', required=False,
                            choices=["tmdev", "dgx", "azureml"], default="tmdev")
    arg_parser.add_argument('--algo_mode', type=str, help='hpo or grid search', required=False,
                            choices=["grid", "hpo", "hfhpo"], default="hpo")
    arg_parser.add_argument('--data_root_dir', type=str, help='data dir', required=False, default="data/")
    arg_parser.add_argument('--dataset_subdataset_name', type=dataset_subdataset_name_format_check,
                            help='dataset and subdataset name', required=False, default=None)
    arg_parser.add_argument('--space_mode', type=str, help='space mode', required=False,
                            choices=["grid", "gnr", "uni", "uni_test", "cus", "buni"], default="uni")
    arg_parser.add_argument('--search_alg_args_mode', type=str, help='search algorithm args mode', required=False,
                            choices=["dft", "exp", "cus"], default="dft")
    arg_parser.add_argument('--algo_name', type=str, help='algorithm', required=False,
                            choices=["bs", "optuna", "cfo", "rs"], default="bs")
    arg_parser.add_argument('--pruner', type=str, help='pruner', required=False,
                            choices=["asha", "None"], default="None")
    arg_parser.add_argument('--pretrained_model_size', type=pretrained_model_size_format_check,
                            help='pretrained model', required=False, default=None)
    arg_parser.add_argument('--sample_num', type=int, help='sample num', required=False, default=None)
    arg_parser.add_argument('--time_budget', type=int, help='time budget', required=False, default=None)
    arg_parser.add_argument('--time_as_grid', type=int, help='time as grid search', required=False, default=None)
    arg_parser.add_argument('--rep_id', type=int, help='rep id', required=False, default=0)
    arg_parser.add_argument('--azure_key', type=str, help='azure key', required=False, default=None)
    arg_parser.add_argument('--resplit_mode', type=str, help='resplit mode', required=False,
                            choices=["rspt", "ori"], default="ori")
    arg_parser.add_argument('--ds_config', type=str, help='deep speed config file path',
                            required=False, default=None)
    arg_parser.add_argument('--yml_file', type=str, help='yml file path', required=False, default="test.yml")
    arg_parser.add_argument('--key_path', type=str, help='path for key.json', required=False, default=None)
    arg_parser.add_argument('--root_log_path', type=str, help='root path for log', required=False, default="logs_azure")
    arg_parser.add_argument('--round_idx', type=int, help='round idx for acl experiments', required=False, default=0)
    arg_parser.add_argument('--seed_data', type=int, help='seed of data shuffling', required=False, default=43)
    arg_parser.add_argument('--seed_transformers', type=int, help='seed of transformers', required=False, default=42)
    console_args, unknown = arg_parser.parse_known_args()
    return console_args


def merge_dicts(dict1, dict2):
    for key2 in dict2.keys():
        if key2 in dict1:
            dict1_vals = set(dict1[key2])
            dict2_vals = set(dict2[key2])
            dict1[key2] = list(dict1_vals.union(dict2_vals))
        else:
            dict1[key2] = dict2[key2]
    return dict1


def _check_dict_keys_overlaps(dict1: dict, dict2: dict):
    dict1_keys = set(dict1.keys())
    dict2_keys = set(dict2.keys())
    return len(dict1_keys.intersection(dict2_keys)) > 0


def _variable_override_default_alternative(obj_ref, var_name, default_value, all_values, overriding_value=None):
    """
        Setting the value of var. If overriding_value is specified, var is set to overriding_value;
        If overriding_value is not specified, var is set to default_value meanwhile showing all_values
    """
    assert isinstance(all_values, list)
    if overriding_value:
        setattr(obj_ref, var_name, overriding_value)
        print("The value for {} is specified as {}".format(var_name, overriding_value))
    else:
        setattr(obj_ref, var_name, default_value)
        print("The value for {} is not specified, setting it to the default value {}. "
              "Alternatively, you can set it to {}".format(var_name, default_value, ",".join(all_values)))


@dataclass
class PathUtils:
    hpo_ckpt_path: str = field(metadata={"help": "the directory for hpo output"})
    hpo_result_path: str = field(metadata={"help": "the directory for hpo result"})
    hpo_log_path: str = field(metadata={"help": "the directory for log"})
    hpo_config_path: str = field(metadata={"help": "the directory for log"})

    log_dir_per_run: str = field(metadata={"help": "log directory for each run."})
    result_dir_per_run: str = field(metadata={"help": "result directory for each run."})
    ckpt_dir_per_run: str = field(metadata={"help": "checkpoint directory for each run."})
    ckpt_dir_per_trial: str = field(metadata={"help": "checkpoint directory for each trial."})

    def __init__(self,
                 jobid_config,
                 hpo_data_root_path,
                 ):
        self.jobid_config = jobid_config
        self.hpo_data_root_path = hpo_data_root_path
        self.hpo_ckpt_path = os.path.join(hpo_data_root_path, "checkpoint")
        self.hpo_result_path = os.path.join(hpo_data_root_path, "result")
        self.hpo_log_path = self.hpo_result_path

    @staticmethod
    def init_and_make_one_dir(dir_path):
        assert dir_path
        if not os.path.exists(dir_path):
            pathlib.Path(dir_path).mkdir(parents=True, exist_ok=True)

    def make_dir_per_run(self):
        jobid_str = self.jobid_config.to_jobid_string()
        self.ckpt_dir_per_run = os.path.join(self.hpo_ckpt_path, jobid_str)
        PathUtils.init_and_make_one_dir(self.ckpt_dir_per_run)

        self.result_dir_per_run = os.path.join(self.hpo_result_path, jobid_str)
        PathUtils.init_and_make_one_dir(self.result_dir_per_run)

        self.log_dir_per_run = os.path.join(self.hpo_log_path, jobid_str)
        PathUtils.init_and_make_one_dir(self.log_dir_per_run)

    def make_dir_per_trial(self, trial_id):
        jobid_str = self.jobid_config.to_jobid_string()
        ckpt_dir_per_run = os.path.join(self.hpo_ckpt_path, jobid_str)
        self.ckpt_dir_per_trial = os.path.join(ckpt_dir_per_run, jobid_str, trial_id)
        PathUtils.init_and_make_one_dir(self.ckpt_dir_per_trial)



import numpy as np
import logging
import time
import math
import copy
import collections
from typing import Optional
from sklearn.metrics import mean_squared_error, mean_absolute_error
from flaml.tune import Trial

logger = logging.getLogger(__name__)


def get_ns_feature_dim_from_vw_example(vw_example) -> dict:
    """Get a dictionary of feature dimensionality for each namespace singleton

    NOTE:
        Assumption: assume the vw_example takes one of the following format
        depending on whether the example includes the feature names

        format 1: 'y |ns1 feature1:feature_value1 feature2:feature_value2 |ns2
                ns2 feature3:feature_value3 feature4:feature_value4'
        format 2: 'y | ns1 feature_value1 feature_value2 |
                ns2 feature_value3 feature_value4'

        The output of both cases are {'ns1': 2, 'ns2': 2}

        For more information about the input formate of vw example, please refer to
        https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format
    """
    ns_feature_dim = {}
    data = vw_example.split('|')
    for i in range(1, len(data)):
        if ':' in data[i]:
            ns_w_feature = data[i].split(' ')
            ns = ns_w_feature[0]
            feature = ns_w_feature[1:]
            feature_dim = len(feature)
        else:
            data_split = data[i].split(' ')
            ns = data_split[0]
            feature_dim = len(data_split) - 1
            if len(data_split[-1]) == 0:
                feature_dim -= 1
        ns_feature_dim[ns] = feature_dim
    logger.debug('name space feature dimension %s', ns_feature_dim)
    return ns_feature_dim


class OnlineResult:
    """Class for managing the result statistics of a trial

    Attributes:
        observation_count: the total number of observations
        resource_used: the sum of loss

    Methods:
        update_result(new_loss, new_resource_used, data_dimension)
            Update result
        get_score(score_name)
            Get the score according to the input score_name
    """
    prob_delta = 0.1
    LOSS_MIN = 0.0
    LOSS_MAX = np.inf
    CB_COEF = 0.05  # 0.001 for mse

    def __init__(self, result_type_name: str, cb_coef: Optional[float] = None,
                 init_loss: Optional[float] = 0.0, init_cb: Optional[float] = 100.0,
                 mode: Optional[str] = 'min', sliding_window_size: Optional[int] = 100):
        """
        Args:
            result_type_name (str): The name of the result type
        """
        self._result_type_name = result_type_name  # for example 'mse' or 'mae'
        self._mode = mode
        self._init_loss = init_loss
        # statistics needed for alg
        self.observation_count = 0
        self.resource_used = 0.0
        self._loss_avg = 0.0
        self._loss_cb = init_cb  # a large number (TODO: this can be changed)
        self._cb_coef = cb_coef if cb_coef is not None else self.CB_COEF
        # optional statistics
        self._sliding_window_size = sliding_window_size
        self._loss_queue = collections.deque(maxlen=self._sliding_window_size)

    def update_result(self, new_loss, new_resource_used, data_dimension,
                      bound_of_range=1.0, new_observation_count=1.0):
        """Update result statistics
        """
        self.resource_used += new_resource_used
        # keep the running average instead of sum of loss to avoid over overflow
        self._loss_avg = self._loss_avg * (self.observation_count / (self.observation_count + new_observation_count)
                                           ) + new_loss / (self.observation_count + new_observation_count)
        self.observation_count += new_observation_count
        self._loss_cb = self._update_loss_cb(bound_of_range, data_dimension)
        self._loss_queue.append(new_loss)

    def _update_loss_cb(self, bound_of_range, data_dim,
                        bound_name='sample_complexity_bound'):
        """Calculate bound coef
        """
        if bound_name == 'sample_complexity_bound':
            # set the coefficient in the loss bound
            if 'mae' in self.result_type_name:
                coef = self._cb_coef * bound_of_range
            else:
                coef = 0.001 * bound_of_range

            comp_F = math.sqrt(data_dim)
            n = self.observation_count
            return coef * comp_F * math.sqrt((np.log10(n / OnlineResult.prob_delta)) / n)
        else:
            raise NotImplementedError

    @property
    def result_type_name(self):
        return self._result_type_name

    @property
    def loss_avg(self):
        return self._loss_avg if \
            self.observation_count != 0 else self._init_loss

    @property
    def loss_cb(self):
        return self._loss_cb

    @property
    def loss_lcb(self):
        return max(self._loss_avg - self._loss_cb, OnlineResult.LOSS_MIN)

    @property
    def loss_ucb(self):
        return min(self._loss_avg + self._loss_cb, OnlineResult.LOSS_MAX)

    @property
    def loss_avg_recent(self):
        return sum(self._loss_queue) / len(self._loss_queue) \
            if len(self._loss_queue) != 0 else self._init_loss

    def get_score(self, score_name, cb_ratio=1):
        if 'lcb' in score_name:
            return max(self._loss_avg - cb_ratio * self._loss_cb, OnlineResult.LOSS_MIN)
        elif 'ucb' in score_name:
            return min(self._loss_avg + cb_ratio * self._loss_cb, OnlineResult.LOSS_MAX)
        elif 'avg' in score_name:
            return self._loss_avg
        else:
            raise NotImplementedError


class BaseOnlineTrial(Trial):
    """Class for online trial.

    Attributes：
        config: the config for this trial
        trial_id: the trial_id of this trial
        min_resource_lease (float): the minimum resource realse
        status: the status of this trial
        start_time: the start time of this trial
        custom_trial_name: a custom name for this trial

    Methods:
        set_resource_lease(resource)
        set_status(status)
        set_checked_under_current_champion(checked_under_current_champion)
    """

    def __init__(self,
                 config: dict,
                 min_resource_lease: float,
                 is_champion: Optional[bool] = False,
                 is_checked_under_current_champion: Optional[bool] = True,
                 custom_trial_name: Optional[str] = 'mae',
                 trial_id: Optional[str] = None,
                 ):
        """
        Args:
            config: the config dict
            min_resource_lease: the minimum resource realse
            is_champion: a bool variable
            is_checked_under_current_champion: a bool variable
            custom_trial_name: custom trial name
            trial_id: the trial id
        """
        # ****basic variables
        self.config = config
        self.trial_id = trial_id
        self.status = Trial.PENDING
        self.start_time = time.time()
        self.custom_trial_name = custom_trial_name

        # ***resource budget related variable
        self._min_resource_lease = min_resource_lease
        self._resource_lease = copy.copy(self._min_resource_lease)
        # ***champion related variables
        self._is_champion = is_champion
        # self._is_checked_under_current_champion_ is supposed to be always 1 when the trial is first created
        self._is_checked_under_current_champion = is_checked_under_current_champion

    @property
    def is_champion(self):
        return self._is_champion

    @property
    def is_checked_under_current_champion(self):
        return self._is_checked_under_current_champion

    @property
    def resource_lease(self):
        return self._resource_lease

    def set_checked_under_current_champion(self, checked_under_current_champion: bool):
        """TODO: add documentation why this is needed. This is needed because sometimes
        we want to know whether a trial has been paused since a new champion is promoted.
        We want to try to pause those running trials (even though they are not yet achieve
        the next scheduling check point according to resource used and resource lease),
        because a better trial is likely to be in the new challengers generated by the new
        champion, so we want to try them as soon as possible.
        If we wait until we reach the next scheduling point, we may waste a lot of resource
        (depending on what is the current resource lease) on the old trials (note that new
        trials is not possible to be scheduled to run until there is a slot openning).
        Intuitively speaking, we want to squize an opening slot as soon as possible once
        a new champion is promoted, such that we are able to try newly generated challengers.
        """
        self._is_checked_under_current_champion = checked_under_current_champion

    def set_resource_lease(self, resource: float):
        self._resource_lease = resource

    def set_status(self, status):
        """Sets the status of the trial and record the start time
        """
        self.status = status
        if status == Trial.RUNNING:
            if self.start_time is None:
                self.start_time = time.time()


class VowpalWabbitTrial(BaseOnlineTrial):
    """Implement BaseOnlineTrial for Vowpal Wabbit

    Attributes:
        model: the online model
        result: the anytime result for the online model
        trainable_class: the model class (set as pyvw.vw for VowpalWabbitTrial)

        config: the config for this trial
        trial_id: the trial_id of this trial
        min_resource_lease (float): the minimum resource realse
        status: the status of this trial
        start_time: the start time of this trial
        custom_trial_name: a custom name for this trial

    Methods:
        set_resource_lease(resource)
        set_status(status)
        set_checked_under_current_champion(checked_under_current_champion)

    NOTE:
        About result:
        1. training related results (need to be updated in the trainable class)
        2. result about resources lease (need to be updated externally)

        About namespaces in vw:
        - Wiki in vw:
        https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Namespaces
        - Namespace vs features:
        https://stackoverflow.com/questions/28586225/in-vowpal-wabbit-what-is-the-difference-between-a-namespace-and-feature
    """
    cost_unit = 1.0
    interactions_config_key = 'interactions'
    MIN_RES_CONST = 5

    def __init__(self,
                 config: dict,
                 min_resource_lease: float,
                 metric: str = 'mae',
                 is_champion: Optional[bool] = False,
                 is_checked_under_current_champion: Optional[bool] = True,
                 custom_trial_name: Optional[str] = 'vw_mae_clipped',
                 trial_id: Optional[str] = None,
                 cb_coef: Optional[float] = None,
                 ):
        """Constructor

        Args:
            config (dict): the config of the trial (note that the config is a set
                because the hyperparameters are )
            min_resource_lease (float): the minimum resource lease
            metric (str): the loss metric
            is_champion (bool): indicates whether the trial is the current champion or not
            is_checked_under_current_champion (bool): indicates whether this trials has
                been paused under the current champion
            trial_id (str): id of the trial (if None, it will be generated in the constructor)

        """
        try:
            from vowpalwabbit import pyvw
        except ImportError:
            raise ImportError(
                'To use AutoVW, please run pip install flaml[vw] to install vowpalwabbit')
        # attributes
        self.trial_id = self._config_to_id(config) if trial_id is None else trial_id
        logger.info('Create trial with trial_id: %s', self.trial_id)
        super().__init__(config, min_resource_lease, is_champion, is_checked_under_current_champion,
                         custom_trial_name, self.trial_id)
        self.model = None   # model is None until the config is scheduled to run
        self.result = None
        self.trainable_class = pyvw.vw
        # variables that are needed during online training
        self._metric = metric
        self._y_min_observed = None
        self._y_max_observed = None
        # application dependent variables
        self._dim = None
        self._cb_coef = cb_coef

    @staticmethod
    def _config_to_id(config):
        """Generate an id for the provided config
        """
        # sort config keys
        sorted_k_list = sorted(list(config.keys()))
        config_id_full = ''
        for key in sorted_k_list:
            v = config[key]
            config_id = '|'
            if isinstance(v, set):
                value_list = sorted(v)
                config_id += '_'.join([str(k) for k in value_list])
            else:
                config_id += str(v)
            config_id_full = config_id_full + config_id
        return config_id_full

    def _initialize_vw_model(self, vw_example):
        """Initialize a vw model using the trainable_class
        """
        self._vw_config = self.config.copy()
        ns_interactions = self.config.get(VowpalWabbitTrial.interactions_config_key, None)
        # ensure the feature interaction config is a list (required by VW)
        if ns_interactions is not None:
            self._vw_config[VowpalWabbitTrial.interactions_config_key] \
                = list(ns_interactions)
        # get the dimensionality of the feature according to the namespace configuration
        namespace_feature_dim = get_ns_feature_dim_from_vw_example(vw_example)
        self._dim = self._get_dim_from_ns(namespace_feature_dim, ns_interactions)
        # construct an instance of vw model using the input config and fixed config
        self.model = self.trainable_class(**self._vw_config)
        self.result = OnlineResult(self._metric,
                                   cb_coef=self._cb_coef,
                                   init_loss=0.0, init_cb=100.0,)

    def train_eval_model_online(self, data_sample, y_pred):
        """Train and eval model online
        """
        # extract info needed the first time we see the data
        if self._resource_lease == 'auto' or self._resource_lease is None:
            assert self._dim is not None
            self._resource_lease = self._dim * self.MIN_RES_CONST
        y = self._get_y_from_vw_example(data_sample)
        self._update_y_range(y)
        if self.model is None:
            # initialize self.model and self.result
            self._initialize_vw_model(data_sample)
        # do one step of learning
        self.model.learn(data_sample)
        # update training related results accordingly
        new_loss = self._get_loss(y, y_pred, self._metric,
                                  self._y_min_observed, self._y_max_observed)
        # udpate sample size, sum of loss, and cost
        data_sample_size = 1
        bound_of_range = self._y_max_observed - self._y_min_observed
        if bound_of_range == 0:
            bound_of_range = 1.0
        self.result.update_result(new_loss,
                                  VowpalWabbitTrial.cost_unit * data_sample_size,
                                  self._dim, bound_of_range)

    def predict(self, x):
        """Predict using the model
        """
        if self.model is None:
            # initialize self.model and self.result
            self._initialize_vw_model(x)
        return self.model.predict(x)

    def _get_loss(self, y_true, y_pred, loss_func_name, y_min_observed, y_max_observed):
        """Get instantaneous loss from y_true and y_pred, and loss_func_name
            For mae_clip, we clip y_pred in the observed range of y
        """
        if 'mse' in loss_func_name or 'squared' in loss_func_name:
            loss_func = mean_squared_error
        elif 'mae' in loss_func_name or 'absolute' in loss_func_name:
            loss_func = mean_absolute_error
            if y_min_observed is not None and y_max_observed is not None and \
               'clip' in loss_func_name:
                # clip y_pred in the observed range of y
                y_pred = min(y_max_observed, max(y_pred, y_min_observed))
        else:
            raise NotImplementedError
        return loss_func([y_true], [y_pred])

    def _update_y_range(self, y):
        """Maintain running observed minimum and maximum target value
        """
        if self._y_min_observed is None or y < self._y_min_observed:
            self._y_min_observed = y
        if self._y_max_observed is None or y > self._y_max_observed:
            self._y_max_observed = y

    @staticmethod
    def _get_dim_from_ns(namespace_feature_dim: dict, namespace_interactions: [set, list]):
        """Get the dimensionality of the corresponding feature of input namespace set
        """
        total_dim = sum(namespace_feature_dim.values())
        if namespace_interactions:
            for f in namespace_interactions:
                ns_dim = 1.0
                for c in f:
                    ns_dim *= namespace_feature_dim[c]
                total_dim += ns_dim
        return total_dim

    def clean_up_model(self):
        self.model = None
        self.result = None

    @staticmethod
    def _get_y_from_vw_example(vw_example):
        """Get y from a vw_example. this works for regression datasets.
        """
        return float(vw_example.split('|')[0])



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License. See LICENSE file in the
 * project root for license information.
'''
from typing import Optional, Union, List, Callable, Tuple
import numpy as np
import datetime
import time
try:
    from ray.tune.analysis import ExperimentAnalysis as EA
except ImportError:
    from .analysis import ExperimentAnalysis as EA
import logging
logger = logging.getLogger(__name__)


_use_ray = True
_runner = None
_verbose = 0
_running_trial = None
_training_iteration = 0


class ExperimentAnalysis(EA):
    '''Class for storing the experiment results
    '''

    def __init__(self, trials, metric, mode):
        try:
            super().__init__(self, None, trials, metric, mode)
        except (TypeError, ValueError):
            self.trials = trials
            self.default_metric = metric or '_default_anonymous_metric'
            self.default_mode = mode


def report(_metric=None, **kwargs):
    '''A function called by the HPO application to report final or intermediate
    results.

    Example:

    .. code-block:: python

        import time
        from flaml import tune

        def compute_with_config(config):
            current_time = time.time()
            metric2minimize = (round(config['x'])-95000)**2
            time2eval = time.time() - current_time
            tune.report(metric2minimize=metric2minimize, time2eval=time2eval)

        analysis = tune.run(
            compute_with_config,
            config={
                'x': tune.lograndint(lower=1, upper=1000000),
                'y': tune.randint(lower=1, upper=1000000)
            },
            metric='metric2minimize', mode='min',
            num_samples=1000000, time_budget_s=60, use_ray=False)

        print(analysis.trials[-1].last_result)

    Args:
        _metric: Optional default anonymous metric for ``tune.report(value)``.
            (For compatibility with ray.tune.report)
        **kwargs: Any key value pair to be reported.
    '''
    global _use_ray
    global _verbose
    global _running_trial
    global _training_iteration
    if _use_ray:
        from ray import tune
        return tune.report(_metric, **kwargs)
    else:
        result = kwargs
        if _verbose == 2:
            logger.info(f"result: {kwargs}")
        if _metric:
            result['_default_anonymous_metric'] = _metric
        trial = _runner.running_trial
        if _running_trial == trial:
            _training_iteration += 1
        else:
            _training_iteration = 0
            _running_trial = trial
        result["training_iteration"] = _training_iteration
        result['config'] = trial.config
        for key, value in trial.config.items():
            result['config/' + key] = value
        _runner.process_trial_result(_runner.running_trial, result)
        result['time_total_s'] = trial.last_update_time - trial.start_time
        if _verbose > 2:
            logger.info(f"result: {result}")
        if _runner.running_trial.is_finished():
            return None
        else:
            return True


def run(training_function,
        config: Optional[dict] = None,
        points_to_evaluate: Optional[List[dict]] = None,
        low_cost_partial_config: Optional[dict] = None,
        cat_hp_cost: Optional[dict] = None,
        metric: Optional[str] = None,
        mode: Optional[str] = None,
        time_budget_s: Union[int, float, datetime.timedelta] = None,
        prune_attr: Optional[str] = None,
        min_resource: Optional[float] = None,
        max_resource: Optional[float] = None,
        reduction_factor: Optional[float] = None,
        report_intermediate_result: Optional[bool] = False,
        search_alg=None,
        verbose: Optional[int] = 2,
        local_dir: Optional[str] = None,
        num_samples: Optional[int] = 1,
        resources_per_trial: Optional[dict] = None,
        config_constraints: Optional[
            List[Tuple[Callable[[dict], float], str, float]]] = None,
        metric_constraints: Optional[
            List[Tuple[str, str, float]]] = None,
        use_ray: Optional[bool] = False):
    '''The trigger for HPO.

    Example:

    .. code-block:: python

        import time
        from flaml import tune

        def compute_with_config(config):
            current_time = time.time()
            metric2minimize = (round(config['x'])-95000)**2
            time2eval = time.time() - current_time
            tune.report(metric2minimize=metric2minimize, time2eval=time2eval)

        analysis = tune.run(
            compute_with_config,
            config={
                'x': tune.lograndint(lower=1, upper=1000000),
                'y': tune.randint(lower=1, upper=1000000)
            },
            metric='metric2minimize', mode='min',
            num_samples=-1, time_budget_s=60, use_ray=False)

        print(analysis.trials[-1].last_result)

    Args:
        training_function: A user-defined training function.
        config: A dictionary to specify the search space.
        points_to_evaluate: A list of initial hyperparameter
            configurations to run first.
        low_cost_partial_config: A dictionary from a subset of
            controlled dimensions to the initial low-cost values.
            e.g.,

            .. code-block:: python

                {'n_estimators': 4, 'max_leaves': 4}

        cat_hp_cost: A dictionary from a subset of categorical dimensions
            to the relative cost of each choice.
            e.g.,

            .. code-block:: python

                {'tree_method': [1, 1, 2]}

            i.e., the relative cost of the
            three choices of 'tree_method' is 1, 1 and 2 respectively
        metric: A string of the metric name to optimize for.
        mode: A string in ['min', 'max'] to specify the objective as
            minimization or maximization.
        time_budget_s: A float of the time budget in seconds.
        prune_attr: A string of the attribute used for pruning.
            Not necessarily in space.
            When prune_attr is in space, it is a hyperparameter, e.g.,
            'n_iters', and the best value is unknown.
            When prune_attr is not in space, it is a resource dimension,
            e.g., 'sample_size', and the peak performance is assumed
            to be at the max_resource.
        min_resource: A float of the minimal resource to use for the
            prune_attr; only valid if prune_attr is not in space.
        max_resource: A float of the maximal resource to use for the
            prune_attr; only valid if prune_attr is not in space.
        reduction_factor: A float of the reduction factor used for incremental
            pruning.
        report_intermediate_result: A boolean of whether intermediate results
            are reported. If so, early stopping and pruning can be used.
        search_alg: An instance of BlendSearch as the search algorithm
            to be used. The same instance can be used for iterative tuning.
            e.g.,

            .. code-block:: python

                from flaml import BlendSearch
                algo = BlendSearch(metric='val_loss', mode='min',
                        space=search_space,
                        low_cost_partial_config=low_cost_partial_config)
                for i in range(10):
                    analysis = tune.run(compute_with_config,
                        search_alg=algo, use_ray=False)
                    print(analysis.trials[-1].last_result)

        verbose: 0, 1, 2, or 3. Verbosity mode for ray if ray backend is used.
            0 = silent, 1 = only status updates, 2 = status and brief trial
            results, 3 = status and detailed trial results. Defaults to 2.
        local_dir: A string of the local dir to save ray logs if ray backend is
            used; or a local dir to save the tuning log.
        num_samples: An integer of the number of configs to try. Defaults to 1.
        resources_per_trial: A dictionary of the hardware resources to allocate
            per trial, e.g., `{'cpu': 1}`. Only valid when using ray backend.
        config_constraints: A list of config constraints to be satisfied.
            e.g.,

            .. code-block: python

                config_constraints = [(mem_size, '<=', 1024**3)]

            mem_size is a function which produces a float number for the bytes
            needed for a config.
            It is used to skip configs which do not fit in memory.
        metric_constraints: A list of metric constraints to be satisfied.
            e.g., `['precision', '>=', 0.9]`
        use_ray: A boolean of whether to use ray as the backend
    '''
    global _use_ray
    global _verbose
    if not use_ray:
        _verbose = verbose
        if verbose > 0:
            import os
            if local_dir:
                os.makedirs(local_dir, exist_ok=True)
                logger.addHandler(logging.FileHandler(local_dir + '/tune_' + str(
                    datetime.datetime.now()).replace(':', '-') + '.log'))
            elif not logger.handlers:
                # Add the console handler.
                _ch = logging.StreamHandler()
                logger_formatter = logging.Formatter(
                    '[%(name)s: %(asctime)s] {%(lineno)d} %(levelname)s - %(message)s',
                    '%m-%d %H:%M:%S')
                _ch.setFormatter(logger_formatter)
                logger.addHandler(_ch)
            if verbose <= 2:
                logger.setLevel(logging.INFO)
            else:
                logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.CRITICAL)

    if search_alg is None:
        from ..searcher.blendsearch import BlendSearch
        search_alg = BlendSearch(
            metric=metric or '_default_anonymous_metric', mode=mode,
            space=config,
            points_to_evaluate=points_to_evaluate,
            low_cost_partial_config=low_cost_partial_config,
            cat_hp_cost=cat_hp_cost,
            prune_attr=prune_attr,
            min_resource=min_resource, max_resource=max_resource,
            reduction_factor=reduction_factor,
            config_constraints=config_constraints,
            metric_constraints=metric_constraints)
    if time_budget_s:
        search_alg.set_search_properties(metric, mode, config={
            'time_budget_s': time_budget_s})
    scheduler = None
    if report_intermediate_result:
        params = {}
        # scheduler resource_dimension=prune_attr
        if prune_attr:
            params['time_attr'] = prune_attr
        if max_resource:
            params['max_t'] = max_resource
        if min_resource:
            params['grace_period'] = min_resource
        if reduction_factor:
            params['reduction_factor'] = reduction_factor
        try:
            from ray.tune.schedulers import ASHAScheduler
            scheduler = ASHAScheduler(**params)
        except ImportError:
            pass
    if use_ray:
        try:
            from ray import tune
        except ImportError:
            raise ImportError("Failed to import ray tune. "
                              "Please install ray[tune] or set use_ray=False")
        _use_ray = True
        return tune.run(training_function,
                        metric=metric, mode=mode,
                        search_alg=search_alg,
                        scheduler=scheduler,
                        time_budget_s=time_budget_s,
                        verbose=verbose, local_dir=local_dir,
                        num_samples=num_samples,
                        resources_per_trial=resources_per_trial)

    # simple sequential run without using tune.run() from ray
    time_start = time.time()
    _use_ray = False
    if scheduler:
        scheduler.set_search_properties(metric=metric, mode=mode)
    from .trial_runner import SequentialTrialRunner
    global _runner
    _runner = SequentialTrialRunner(
        search_alg=search_alg,
        scheduler=scheduler,
        metric=metric,
        mode=mode,
    )
    num_trials = 0
    if time_budget_s is None:
        time_budget_s = np.inf
    while time.time() - time_start < time_budget_s and (
            num_samples < 0 or num_trials < num_samples):
        trial_to_run = _runner.step()
        if trial_to_run:
            num_trials += 1
            if verbose:
                logger.info(f'trial {num_trials} config: {trial_to_run.config}')
            result = training_function(trial_to_run.config)
            if result is not None:
                if isinstance(result, dict):
                    tune.report(**result)
                else:
                    tune.report(_metric=result)
            _runner.stop_trial(trial_to_run)
    if verbose > 0:
        logger.handlers.clear()
    return ExperimentAnalysis(_runner.get_trials(), metric=metric, mode=mode)



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
'''

import json
from typing import IO
from contextlib import contextmanager
import warnings


class TrainingLogRecord(object):

    def __init__(self,
                 record_id: int,
                 iter_per_learner: int,
                 logged_metric: float,
                 trial_time: float,
                 total_search_time: float,
                 validation_loss,
                 config,
                 best_validation_loss,
                 best_config,
                 learner,
                 sample_size):
        self.record_id = record_id
        self.iter_per_learner = iter_per_learner
        self.logged_metric = logged_metric
        self.trial_time = trial_time
        self.total_search_time = total_search_time
        self.validation_loss = validation_loss
        self.config = config
        self.best_validation_loss = best_validation_loss
        self.best_config = best_config
        self.learner = learner
        self.sample_size = sample_size

    def dump(self, fp: IO[str]):
        d = vars(self)
        return json.dump(d, fp)

    @classmethod
    def load(cls, json_str: str):
        d = json.loads(json_str)
        return cls(**d)

    def __str__(self):
        return json.dumps(vars(self))


class TrainingLogCheckPoint(TrainingLogRecord):

    def __init__(self, curr_best_record_id: int):
        self.curr_best_record_id = curr_best_record_id


class TrainingLogWriter(object):

    def __init__(self, output_filename: str):
        self.output_filename = output_filename
        self.file = None
        self.current_best_loss_record_id = None
        self.current_best_loss = float('+inf')
        self.current_sample_size = None
        self.current_record_id = 0

    def open(self):
        self.file = open(self.output_filename, 'w')

    def append(self,
               it_counter: int,
               train_loss: float,
               trial_time: float,
               total_search_time: float,
               validation_loss,
               config,
               best_validation_loss,
               best_config,
               learner,
               sample_size):
        if self.file is None:
            raise IOError("Call open() to open the outpute file first.")
        if validation_loss is None:
            raise ValueError('TEST LOSS NONE ERROR!!!')
        record = TrainingLogRecord(self.current_record_id,
                                   it_counter,
                                   train_loss,
                                   trial_time,
                                   total_search_time,
                                   validation_loss,
                                   config,
                                   best_validation_loss,
                                   best_config,
                                   learner,
                                   sample_size)
        if validation_loss < self.current_best_loss or \
            validation_loss == self.current_best_loss and \
                sample_size > self.current_sample_size:
            self.current_best_loss = validation_loss
            self.current_sample_size = sample_size
            self.current_best_loss_record_id = self.current_record_id
        self.current_record_id += 1
        record.dump(self.file)
        self.file.write('\n')
        self.file.flush()

    def checkpoint(self):
        if self.file is None:
            raise IOError("Call open() to open the outpute file first.")
        if self.current_best_loss_record_id is None:
            warnings.warn("checkpoint() called before any record is written, "
                          "skipped.")
            return
        record = TrainingLogCheckPoint(self.current_best_loss_record_id)
        record.dump(self.file)
        self.file.write('\n')
        self.file.flush()

    def close(self):
        self.file.close()
        self.file = None  # for pickle


class TrainingLogReader(object):

    def __init__(self, filename: str):
        self.filename = filename
        self.file = None

    def open(self):
        self.file = open(self.filename)

    def records(self):
        if self.file is None:
            raise IOError("Call open() before reading log file.")
        for line in self.file:
            data = json.loads(line)
            if len(data) == 1:
                # Skip checkpoints.
                continue
            yield TrainingLogRecord(**data)

    def close(self):
        self.file.close()
        self.file = None  # for pickle

    def get_record(self, record_id) -> TrainingLogRecord:
        if self.file is None:
            raise IOError("Call open() before reading log file.")
        for rec in self.records():
            if rec.record_id == record_id:
                return rec
        raise ValueError(f"Cannot find record with id {record_id}.")


@contextmanager
def training_log_writer(filename: str):
    try:
        w = TrainingLogWriter(filename)
        w.open()
        yield w
    finally:
        w.close()


@contextmanager
def training_log_reader(filename: str):
    try:
        r = TrainingLogReader(filename)
        r.open()
        yield r
    finally:
        r.close()



from flaml.searcher import CFO, BlendSearch, FLOW2, BlendSearchTuner
from flaml.automl import AutoML, logger_formatter
from flaml.onlineml.autovw import AutoVW
from flaml.version import __version__
import logging

# Set the root logger.
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)



'''!
 * Copyright (c) 2020-2021 Microsoft Corporation. All rights reserved.
 * Licensed under the MIT License.
'''

import time
from joblib.externals.cloudpickle.cloudpickle import instance
import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, r2_score, roc_auc_score, \
    accuracy_score, mean_absolute_error, log_loss, average_precision_score, \
    f1_score
from sklearn.model_selection import RepeatedStratifiedKFold, GroupKFold
from .model import (
    XGBoostEstimator, XGBoostSklearnEstimator, RandomForestEstimator,
    LGBMEstimator, LRL1Classifier, LRL2Classifier, CatBoostEstimator,
    ExtraTreeEstimator, KNeighborsEstimator)

import logging
logger = logging.getLogger(__name__)


def get_estimator_class(task, estimator_name):
    ''' when adding a new learner, need to add an elif branch '''

    if 'xgboost' in estimator_name:
        if 'regression' in task:
            estimator_class = XGBoostEstimator
        else:
            estimator_class = XGBoostSklearnEstimator
    elif 'rf' in estimator_name:
        estimator_class = RandomForestEstimator
    elif 'lgbm' in estimator_name:
        estimator_class = LGBMEstimator
    elif 'lrl1' in estimator_name:
        estimator_class = LRL1Classifier
    elif 'lrl2' in estimator_name:
        estimator_class = LRL2Classifier
    elif 'catboost' in estimator_name:
        estimator_class = CatBoostEstimator
    elif 'extra_tree' in estimator_name:
        estimator_class = ExtraTreeEstimator
    elif 'kneighbor' in estimator_name:
        estimator_class = KNeighborsEstimator
    else:
        raise ValueError(
            estimator_name + ' is not a built-in learner. '
            'Please use AutoML.add_learner() to add a customized learner.')
    return estimator_class


def sklearn_metric_loss_score(
    metric_name, y_predict, y_true, labels=None, sample_weight=None
):
    '''Loss using the specified metric

    Args:
        metric_name: A string of the metric name, one of
            'r2', 'rmse', 'mae', 'mse', 'accuracy', 'roc_auc', 'log_loss',
            'f1', 'ap', 'micro_f1', 'macro_f1'
        y_predict: A 1d or 2d numpy array of the predictions which can be
            used to calculate the metric. E.g., 2d for log_loss and 1d
            for others.
        y_true: A 1d numpy array of the true labels
        labels: A 1d numpy array of the unique labels
        sample_weight: A 1d numpy array of the sample weight

    Returns:
        score: A float number of the loss, the lower the better
    '''
    metric_name = metric_name.lower()
    if 'r2' in metric_name:
        score = 1.0 - r2_score(y_true, y_predict, sample_weight=sample_weight)
    elif metric_name == 'rmse':
        score = np.sqrt(mean_squared_error(
            y_true, y_predict, sample_weight=sample_weight))
    elif metric_name == 'mae':
        score = mean_absolute_error(
            y_true, y_predict, sample_weight=sample_weight)
    elif metric_name == 'mse':
        score = mean_squared_error(
            y_true, y_predict, sample_weight=sample_weight)
    elif metric_name == 'accuracy':
        score = 1.0 - accuracy_score(
            y_true, y_predict, sample_weight=sample_weight)
    elif 'roc_auc' in metric_name:
        score = 1.0 - roc_auc_score(
            y_true, y_predict, sample_weight=sample_weight)
    elif 'log_loss' in metric_name:
        score = log_loss(
            y_true, y_predict, labels=labels, sample_weight=sample_weight)
    elif 'micro_f1' in metric_name:
        score = 1 - f1_score(
            y_true, y_predict, sample_weight=sample_weight, average='micro')
    elif 'macro_f1' in metric_name:
        score = 1 - f1_score(
            y_true, y_predict, sample_weight=sample_weight, average='macro')
    elif 'f1' in metric_name:
        score = 1 - f1_score(y_true, y_predict, sample_weight=sample_weight)
    elif 'ap' in metric_name:
        score = 1 - average_precision_score(
            y_true, y_predict, sample_weight=sample_weight)
    else:
        raise ValueError(
            metric_name + ' is not a built-in metric, '
            'currently built-in metrics are: '
            'r2, rmse, mae, mse, accuracy, roc_auc, log_loss, f1, micro_f1, macro_f1, ap. '
            'please pass a customized metric function to AutoML.fit(metric=func)')
    return score


def get_y_pred(estimator, X, eval_metric, obj):
    if eval_metric in ['roc_auc', 'ap'] and 'binary' in obj:
        y_pred_classes = estimator.predict_proba(X)
        y_pred = y_pred_classes[
            :, 1] if y_pred_classes.ndim > 1 else y_pred_classes
    elif eval_metric in ['log_loss', 'roc_auc']:
        y_pred = estimator.predict_proba(X)
    else:
        y_pred = estimator.predict(X)
    return y_pred


def get_test_loss(
    estimator, X_train, y_train, X_test, y_test, weight_test,
    eval_metric, obj, labels=None, budget=None, train_loss=False, fit_kwargs={}
):
    start = time.time()
    train_time = estimator.fit(X_train, y_train, budget, **fit_kwargs)
    if isinstance(eval_metric, str):
        pred_start = time.time()
        test_pred_y = get_y_pred(estimator, X_test, eval_metric, obj)
        pred_time = (time.time() - pred_start) / X_test.shape[0]
        test_loss = sklearn_metric_loss_score(eval_metric, test_pred_y, y_test,
                                              labels, weight_test)
        if train_loss is not False:
            test_pred_y = get_y_pred(estimator, X_train, eval_metric, obj)
            train_loss = sklearn_metric_loss_score(
                eval_metric, test_pred_y,
                y_train, labels, fit_kwargs.get('sample_weight'))
    else:  # customized metric function
        test_loss, metrics = eval_metric(
            X_test, y_test, estimator, labels, X_train, y_train,
            weight_test, fit_kwargs.get('sample_weight'))
        if isinstance(metrics, dict):
            pred_time = metrics.get('pred_time', 0)
        train_loss = metrics
    train_time = time.time() - start
    return test_loss, train_time, train_loss, pred_time


def train_model(estimator, X_train, y_train, budget, fit_kwargs={}):
    train_time = estimator.fit(X_train, y_train, budget, **fit_kwargs)
    return train_time


def evaluate_model(
    estimator, X_train, y_train, X_val, y_val, weight_val,
    budget, kf, task, eval_method, eval_metric, best_val_loss, train_loss=False,
    fit_kwargs={}
):
    if 'holdout' in eval_method:
        val_loss, train_loss, train_time, pred_time = evaluate_model_holdout(
            estimator, X_train, y_train, X_val, y_val, weight_val, budget,
            task, eval_metric, train_loss=train_loss,
            fit_kwargs=fit_kwargs)
    else:
        val_loss, train_loss, train_time, pred_time = evaluate_model_CV(
            estimator, X_train, y_train, budget, kf, task,
            eval_metric, best_val_loss, train_loss=train_loss,
            fit_kwargs=fit_kwargs)
    return val_loss, train_loss, train_time, pred_time


def evaluate_model_holdout(
    estimator, X_train, y_train, X_val, y_val,
    weight_val, budget, task, eval_metric, train_loss=False,
    fit_kwargs={}
):
    val_loss, train_time, train_loss, pred_time = get_test_loss(
        estimator, X_train, y_train, X_val, y_val, weight_val, eval_metric,
        task, budget=budget, train_loss=train_loss, fit_kwargs=fit_kwargs)
    return val_loss, train_loss, train_time, pred_time


def evaluate_model_CV(
    estimator, X_train_all, y_train_all, budget, kf,
    task, eval_metric, best_val_loss, train_loss=False, fit_kwargs={}
):
    start_time = time.time()
    total_val_loss = 0
    total_train_loss = None
    train_time = pred_time = 0
    valid_fold_num = total_fold_num = 0
    n = kf.get_n_splits()
    X_train_split, y_train_split = X_train_all, y_train_all
    if task == 'regression':
        labels = None
    else:
        labels = np.unique(y_train_all)

    if isinstance(kf, RepeatedStratifiedKFold):
        kf = kf.split(X_train_split, y_train_split)
    elif isinstance(kf, GroupKFold):
        kf = kf.split(X_train_split, y_train_split, kf.groups)
    else:
        kf = kf.split(X_train_split)
    rng = np.random.RandomState(2020)
    val_loss_list = []
    budget_per_train = budget / (n + 1)
    if 'sample_weight' in fit_kwargs:
        weight = fit_kwargs['sample_weight']
        weight_val = None
    else:
        weight = weight_val = None
    for train_index, val_index in kf:
        train_index = rng.permutation(train_index)
        if isinstance(X_train_all, pd.DataFrame):
            X_train, X_val = X_train_split.iloc[
                train_index], X_train_split.iloc[val_index]
        else:
            X_train, X_val = X_train_split[
                train_index], X_train_split[val_index]
        if isinstance(y_train_all, pd.Series):
            y_train, y_val = y_train_split.iloc[
                train_index], y_train_split.iloc[val_index]
        else:
            y_train, y_val = y_train_split[
                train_index], y_train_split[val_index]
        estimator.cleanup()
        if weight is not None:
            fit_kwargs['sample_weight'], weight_val = weight[
                train_index], weight[val_index]
        val_loss_i, train_time_i, train_loss_i, pred_time_i = get_test_loss(
            estimator, X_train, y_train, X_val, y_val, weight_val,
            eval_metric, task, labels, budget_per_train,
            train_loss=train_loss, fit_kwargs=fit_kwargs)
        if weight is not None:
            fit_kwargs['sample_weight'] = weight
        valid_fold_num += 1
        total_fold_num += 1
        total_val_loss += val_loss_i
        if train_loss is not False:
            if isinstance(total_train_loss, list):
                total_train_loss = [
                    total_train_loss[i] + v for i, v in enumerate(train_loss_i)]
            elif isinstance(total_train_loss, dict):
                total_train_loss = {
                    k: total_train_loss[k] + v for k, v in train_loss_i.items()}
            elif total_train_loss is not None:
                total_train_loss += train_loss_i
            else:
                total_train_loss = train_loss_i
        train_time += train_time_i
        pred_time += pred_time_i
        if valid_fold_num == n:
            val_loss_list.append(total_val_loss / valid_fold_num)
            total_val_loss = valid_fold_num = 0
        elif time.time() - start_time >= budget:
            val_loss_list.append(total_val_loss / valid_fold_num)
            break
    val_loss = np.max(val_loss_list)
    n = total_fold_num
    if train_loss is not False:
        if isinstance(total_train_loss, list):
            train_loss = [v / n for v in total_train_loss]
        elif isinstance(total_train_loss, dict):
            train_loss = {k: v / n for k, v in total_train_loss.items()}
        else:
            train_loss = total_train_loss / n
    pred_time /= n
    budget -= time.time() - start_time
    if val_loss < best_val_loss and budget > budget_per_train:
        estimator.cleanup()
        estimator.fit(X_train_all, y_train_all, budget, **fit_kwargs)
    return val_loss, train_loss, train_time, pred_time


def compute_estimator(
    X_train, y_train, X_val, y_val, weight_val, budget, kf,
    config_dic, task, estimator_name, eval_method, eval_metric,
    best_val_loss=np.Inf, n_jobs=1, estimator_class=None, train_loss=False,
    fit_kwargs={}
):
    estimator_class = estimator_class or get_estimator_class(
        task, estimator_name)
    estimator = estimator_class(
        **config_dic, task=task, n_jobs=n_jobs)
    val_loss, train_loss, train_time, pred_time = evaluate_model(
        estimator, X_train, y_train, X_val, y_val, weight_val, budget, kf, task,
        eval_method, eval_metric, best_val_loss, train_loss=train_loss,
        fit_kwargs=fit_kwargs)
    return estimator, val_loss, train_loss, train_time, pred_time


def train_estimator(
    X_train, y_train, config_dic, task,
    estimator_name, n_jobs=1, estimator_class=None, budget=None, fit_kwargs={}
):
    start_time = time.time()
    estimator_class = estimator_class or get_estimator_class(
        task, estimator_name)
    estimator = estimator_class(**config_dic, task=task, n_jobs=n_jobs)
    if X_train is not None:
        train_time = train_model(
            estimator, X_train, y_train, budget, fit_kwargs)
    else:
        estimator = estimator.estimator_class(**estimator.params)
    train_time = time.time() - start_time
    return estimator, train_time


def get_classification_objective(num_labels: int) -> str:
    if num_labels == 2:
        objective_name = 'binary:logistic'
    else:
        objective_name = 'multi:softmax'
    return objective_name


def norm_confusion_matrix(y_true, y_pred):
    '''normalized confusion matrix

    Args:
        estimator: A multi-class classification estimator
        y_true: A numpy array or a pandas series of true labels
        y_pred: A numpy array or a pandas series of predicted labels

    Returns:
        A normalized confusion matrix
    '''
    from sklearn.metrics import confusion_matrix
    conf_mat = confusion_matrix(y_true, y_pred)
    norm_conf_mat = conf_mat.astype('float') / conf_mat.sum(axis=1)[:, np.newaxis]
    return norm_conf_mat


def multi_class_curves(y_true, y_pred_proba, curve_func):
    '''Binarize the data for multi-class tasks and produce ROC or precision-recall curves

    Args:
        y_true: A numpy array or a pandas series of true labels
        y_pred_proba: A numpy array or a pandas dataframe of predicted probabilites
        curve_func: A function to produce a curve (e.g., roc_curve or precision_recall_curve)

    Returns:
        A tuple of two dictionaries with the same set of keys (class indices)
        The first dictionary curve_x stores the x coordinates of each curve, e.g.,
            curve_x[0] is an 1D array of the x coordinates of class 0
        The second dictionary curve_y stores the y coordinates of each curve, e.g.,
            curve_y[0] is an 1D array of the y coordinates of class 0
    '''
    from sklearn.preprocessing import label_binarize
    classes = np.unique(y_true)
    y_true_binary = label_binarize(y_true, classes=classes)

    curve_x, curve_y = {}, {}
    for i in range(len(classes)):
        curve_x[i], curve_y[i], _ = curve_func(y_true_binary[:, i], y_pred_proba[:, i])
    return curve_x, curve_y



import setuptools
import os

here = os.path.abspath(os.path.dirname(__file__))

with open("README.md", "r") as fh:
    long_description = fh.read()


# Get the code version
version = {}
with open(os.path.join(here, "flaml/version.py")) as fp:
    exec(fp.read(), version)
__version__ = version["__version__"]

install_requires = [
    "NumPy>=1.16.2",
    "lightgbm>=2.3.1",
    "xgboost>=0.90",
    "scipy>=1.4.1",
    "catboost>=0.23",
    "scikit-learn>=0.23.2",
],


setuptools.setup(
    name="FLAML",
    version=__version__,
    author="Microsoft Corporation",
    author_email="hpo@microsoft.com",
    description="A fast and lightweight autoML system",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/FLAML",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    extras_require={
        "notebook": [
            "openml==0.10.2",
            "jupyter",
            "matplotlib==3.2.0",
            "rgf-python",
            "vowpalwabbit",
        ],
        "test": [
            "flake8>=3.8.4",
            "pytest>=6.1.1",
            "coverage>=5.3",
            "xgboost<1.3",
            "rgf-python",
            "optuna==2.8.0",
            "vowpalwabbit",
            "openml",
            "transformers==4.4.1",
            "wandb==0.10.26",
            "torch==1.8.1",
            "datasets==1.4.1",
            "azure-storage-blob",
        ],
        "blendsearch": [
            "optuna==2.8.0"
        ],
        "ray": [
            "ray[tune]==1.4.1",
            "pyyaml<5.3.1",
        ],
        "azureml": [
            "azureml-mlflow",
        ],
        "nni": [
            "nni",
        ],
        "vw": [
            "vowpalwabbit",
        ],
        "nlp": [
            "ray[tune]>=1.4.1",
            "transformers",
            "datasets==1.4.1",
            "tensorboardX<=2.2",
            "torch"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



'''
Copyright 2020 The Ray Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This source file is included here because ray does not fully support Windows.

Copyright (c) Microsoft Corporation.
'''
import logging
import random
from copy import copy
from inspect import signature
from math import isclose
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import numpy as np

logger = logging.getLogger(__name__)


class Domain:
    """Base class to specify a type and valid range to sample parameters from.
    This base class is implemented by parameter spaces, like float ranges
    (``Float``), integer ranges (``Integer``), or categorical variables
    (``Categorical``). The ``Domain`` object contains information about
    valid values (e.g. minimum and maximum values), and exposes methods that
    allow specification of specific samplers (e.g. ``uniform()`` or
    ``loguniform()``).
    """
    sampler = None
    default_sampler_cls = None

    def cast(self, value):
        """Cast value to domain type"""
        return value

    def set_sampler(self, sampler, allow_override=False):
        if self.sampler and not allow_override:
            raise ValueError("You can only choose one sampler for parameter "
                             "domains. Existing sampler for parameter {}: "
                             "{}. Tried to add {}".format(
                                 self.__class__.__name__, self.sampler,
                                 sampler))
        self.sampler = sampler

    def get_sampler(self):
        sampler = self.sampler
        if not sampler:
            sampler = self.default_sampler_cls()
        return sampler

    def sample(self, spec=None, size=1):
        sampler = self.get_sampler()
        return sampler.sample(self, spec=spec, size=size)

    def is_grid(self):
        return isinstance(self.sampler, Grid)

    def is_function(self):
        return False

    def is_valid(self, value: Any):
        """Returns True if `value` is a valid value in this domain."""
        raise NotImplementedError

    @property
    def domain_str(self):
        return "(unknown)"


class Sampler:
    def sample(self,
               domain: Domain,
               spec: Optional[Union[List[Dict], Dict]] = None,
               size: int = 1):
        raise NotImplementedError


class BaseSampler(Sampler):
    def __str__(self):
        return "Base"


class Uniform(Sampler):
    def __str__(self):
        return "Uniform"


class LogUniform(Sampler):
    def __init__(self, base: float = 10):
        self.base = base
        assert self.base > 0, "Base has to be strictly greater than 0"

    def __str__(self):
        return "LogUniform"


class Normal(Sampler):
    def __init__(self, mean: float = 0., sd: float = 0.):
        self.mean = mean
        self.sd = sd

        assert self.sd > 0, "SD has to be strictly greater than 0"

    def __str__(self):
        return "Normal"


class Grid(Sampler):
    """Dummy sampler used for grid search"""

    def sample(self,
               domain: Domain,
               spec: Optional[Union[List[Dict], Dict]] = None,
               size: int = 1):
        return RuntimeError("Do not call `sample()` on grid.")


class Float(Domain):
    class _Uniform(Uniform):
        def sample(self,
                   domain: "Float",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            assert domain.lower > float("-inf"), \
                "Uniform needs a lower bound"
            assert domain.upper < float("inf"), \
                "Uniform needs a upper bound"
            items = np.random.uniform(domain.lower, domain.upper, size=size)
            return items if len(items) > 1 else domain.cast(items[0])

    class _LogUniform(LogUniform):
        def sample(self,
                   domain: "Float",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            assert domain.lower > 0, \
                "LogUniform needs a lower bound greater than 0"
            assert 0 < domain.upper < float("inf"), \
                "LogUniform needs a upper bound greater than 0"
            logmin = np.log(domain.lower) / np.log(self.base)
            logmax = np.log(domain.upper) / np.log(self.base)

            items = self.base**(np.random.uniform(logmin, logmax, size=size))
            return items if len(items) > 1 else domain.cast(items[0])

    class _Normal(Normal):
        def sample(self,
                   domain: "Float",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            assert not domain.lower or domain.lower == float("-inf"), \
                "Normal sampling does not allow a lower value bound."
            assert not domain.upper or domain.upper == float("inf"), \
                "Normal sampling does not allow a upper value bound."
            items = np.random.normal(self.mean, self.sd, size=size)
            return items if len(items) > 1 else domain.cast(items[0])

    default_sampler_cls = _Uniform

    def __init__(self, lower: Optional[float], upper: Optional[float]):
        # Need to explicitly check for None
        self.lower = lower if lower is not None else float("-inf")
        self.upper = upper if upper is not None else float("inf")

    def cast(self, value):
        return float(value)

    def uniform(self):
        if not self.lower > float("-inf"):
            raise ValueError(
                "Uniform requires a lower bound. Make sure to set the "
                "`lower` parameter of `Float()`.")
        if not self.upper < float("inf"):
            raise ValueError(
                "Uniform requires a upper bound. Make sure to set the "
                "`upper` parameter of `Float()`.")
        new = copy(self)
        new.set_sampler(self._Uniform())
        return new

    def loguniform(self, base: float = 10):
        if not self.lower > 0:
            raise ValueError(
                "LogUniform requires a lower bound greater than 0."
                f"Got: {self.lower}. Did you pass a variable that has "
                "been log-transformed? If so, pass the non-transformed value "
                "instead.")
        if not 0 < self.upper < float("inf"):
            raise ValueError(
                "LogUniform requires a upper bound greater than 0. "
                f"Got: {self.lower}. Did you pass a variable that has "
                "been log-transformed? If so, pass the non-transformed value "
                "instead.")
        new = copy(self)
        new.set_sampler(self._LogUniform(base))
        return new

    def normal(self, mean=0., sd=1.):
        new = copy(self)
        new.set_sampler(self._Normal(mean, sd))
        return new

    def quantized(self, q: float):
        if self.lower > float("-inf") and not isclose(self.lower / q,
                                                      round(self.lower / q)):
            raise ValueError(
                f"Your lower variable bound {self.lower} is not divisible by "
                f"quantization factor {q}.")
        if self.upper < float("inf") and not isclose(self.upper / q,
                                                     round(self.upper / q)):
            raise ValueError(
                f"Your upper variable bound {self.upper} is not divisible by "
                f"quantization factor {q}.")

        new = copy(self)
        new.set_sampler(Quantized(new.get_sampler(), q), allow_override=True)
        return new

    def is_valid(self, value: float):
        return self.lower <= value <= self.upper

    @property
    def domain_str(self):
        return f"({self.lower}, {self.upper})"


class Integer(Domain):
    class _Uniform(Uniform):
        def sample(self,
                   domain: "Integer",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            items = np.random.randint(domain.lower, domain.upper, size=size)
            return items if len(items) > 1 else domain.cast(items[0])

    class _LogUniform(LogUniform):
        def sample(self,
                   domain: "Integer",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            assert domain.lower > 0, \
                "LogUniform needs a lower bound greater than 0"
            assert 0 < domain.upper < float("inf"), \
                "LogUniform needs a upper bound greater than 0"
            logmin = np.log(domain.lower) / np.log(self.base)
            logmax = np.log(domain.upper) / np.log(self.base)

            items = self.base**(np.random.uniform(logmin, logmax, size=size))
            items = np.round(items).astype(int)
            return items if len(items) > 1 else domain.cast(items[0])

    default_sampler_cls = _Uniform

    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper

    def cast(self, value):
        return int(value)

    def quantized(self, q: int):
        new = copy(self)
        new.set_sampler(Quantized(new.get_sampler(), q), allow_override=True)
        return new

    def uniform(self):
        new = copy(self)
        new.set_sampler(self._Uniform())
        return new

    def loguniform(self, base: float = 10):
        if not self.lower > 0:
            raise ValueError(
                "LogUniform requires a lower bound greater than 0."
                f"Got: {self.lower}. Did you pass a variable that has "
                "been log-transformed? If so, pass the non-transformed value "
                "instead.")
        if not 0 < self.upper < float("inf"):
            raise ValueError(
                "LogUniform requires a upper bound greater than 0. "
                f"Got: {self.lower}. Did you pass a variable that has "
                "been log-transformed? If so, pass the non-transformed value "
                "instead.")
        new = copy(self)
        new.set_sampler(self._LogUniform(base))
        return new

    def is_valid(self, value: int):
        return self.lower <= value <= self.upper

    @property
    def domain_str(self):
        return f"({self.lower}, {self.upper})"


class Categorical(Domain):
    class _Uniform(Uniform):
        def sample(self,
                   domain: "Categorical",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):

            items = random.choices(domain.categories, k=size)
            return items if len(items) > 1 else domain.cast(items[0])

    default_sampler_cls = _Uniform

    def __init__(self, categories: Sequence):
        self.categories = list(categories)

    def uniform(self):
        new = copy(self)
        new.set_sampler(self._Uniform())
        return new

    def grid(self):
        new = copy(self)
        new.set_sampler(Grid())
        return new

    def __len__(self):
        return len(self.categories)

    def __getitem__(self, item):
        return self.categories[item]

    def is_valid(self, value: Any):
        return value in self.categories

    @property
    def domain_str(self):
        return f"{self.categories}"


class Function(Domain):
    class _CallSampler(BaseSampler):
        def sample(self,
                   domain: "Function",
                   spec: Optional[Union[List[Dict], Dict]] = None,
                   size: int = 1):
            if domain.pass_spec:
                items = [
                    domain.func(spec[i] if isinstance(spec, list) else spec)
                    for i in range(size)
                ]
            else:
                items = [domain.func() for i in range(size)]

            return items if len(items) > 1 else domain.cast(items[0])

    default_sampler_cls = _CallSampler

    def __init__(self, func: Callable):
        sig = signature(func)

        pass_spec = True  # whether we should pass `spec` when calling `func`
        try:
            sig.bind({})
        except TypeError:
            pass_spec = False

        if not pass_spec:
            try:
                sig.bind()
            except TypeError as exc:
                raise ValueError(
                    "The function passed to a `Function` parameter must be "
                    "callable with either 0 or 1 parameters.") from exc

        self.pass_spec = pass_spec
        self.func = func

    def is_function(self):
        return True

    def is_valid(self, value: Any):
        return True  # This is user-defined, so lets not assume anything

    @property
    def domain_str(self):
        return f"{self.func}()"


class Quantized(Sampler):
    def __init__(self, sampler: Sampler, q: Union[float, int]):
        self.sampler = sampler
        self.q = q

        assert self.sampler, "Quantized() expects a sampler instance"

    def get_sampler(self):
        return self.sampler

    def sample(self,
               domain: Domain,
               spec: Optional[Union[List[Dict], Dict]] = None,
               size: int = 1):
        values = self.sampler.sample(domain, spec, size)
        quantized = np.round(np.divide(values, self.q)) * self.q
        if not isinstance(quantized, np.ndarray):
            return domain.cast(quantized)
        return list(quantized)


class PolynomialExpansionSet:

    def __init__(self, init_monomials: set = (), highest_poly_order: int = None,
                 allow_self_inter: bool = False):
        self._init_monomials = init_monomials
        self._highest_poly_order = highest_poly_order if \
            highest_poly_order is not None else len(self._init_monomials)
        self._allow_self_inter = allow_self_inter

    @property
    def init_monomials(self):
        return self._init_monomials

    @property
    def highest_poly_order(self):
        return self._highest_poly_order

    @property
    def allow_self_inter(self):
        return self._allow_self_inter

    def __str__(self):
        return "PolynomialExpansionSet"


# TODO (krfricke): Remove tune.function
def function(func):
    logger.warning(
        "DeprecationWarning: wrapping {} with tune.function() is no "
        "longer needed".format(func))
    return func


def sample_from(func: Callable[[Dict], Any]):
    """Specify that tune should sample configuration values from this function.
    Arguments:
        func: An callable function to draw a sample from.
    """
    return Function(func)


def uniform(lower: float, upper: float):
    """Sample a float value uniformly between ``lower`` and ``upper``.
    Sampling from ``tune.uniform(1, 10)`` is equivalent to sampling from
    ``np.random.uniform(1, 10))``
    """
    return Float(lower, upper).uniform()


def quniform(lower: float, upper: float, q: float):
    """Sample a quantized float value uniformly between ``lower`` and ``upper``.
    Sampling from ``tune.uniform(1, 10)`` is equivalent to sampling from
    ``np.random.uniform(1, 10))``
    The value will be quantized, i.e. rounded to an integer increment of ``q``.
    Quantization makes the upper bound inclusive.
    """
    return Float(lower, upper).uniform().quantized(q)


def loguniform(lower: float, upper: float, base: float = 10):
    """Sugar for sampling in different orders of magnitude.
    Args:
        lower (float): Lower boundary of the output interval (e.g. 1e-4)
        upper (float): Upper boundary of the output interval (e.g. 1e-2)
        base (int): Base of the log. Defaults to 10.
    """
    return Float(lower, upper).loguniform(base)


def qloguniform(lower: float, upper: float, q: float, base: float = 10):
    """Sugar for sampling in different orders of magnitude.
    The value will be quantized, i.e. rounded to an integer increment of ``q``.
    Quantization makes the upper bound inclusive.
    Args:
        lower (float): Lower boundary of the output interval (e.g. 1e-4)
        upper (float): Upper boundary of the output interval (e.g. 1e-2)
        q (float): Quantization number. The result will be rounded to an
            integer increment of this value.
        base (int): Base of the log. Defaults to 10.
    """
    return Float(lower, upper).loguniform(base).quantized(q)


def choice(categories: List):
    """Sample a categorical value.
    Sampling from ``tune.choice([1, 2])`` is equivalent to sampling from
    ``random.choice([1, 2])``
    """
    return Categorical(categories).uniform()


def randint(lower: int, upper: int):
    """Sample an integer value uniformly between ``lower`` and ``upper``.
    ``lower`` is inclusive, ``upper`` is exclusive.
    Sampling from ``tune.randint(10)`` is equivalent to sampling from
    ``np.random.randint(10)``
    """
    return Integer(lower, upper).uniform()


def lograndint(lower: int, upper: int, base: float = 10):
    """Sample an integer value log-uniformly between ``lower`` and ``upper``,
    with ``base`` being the base of logarithm.
    ``lower`` is inclusive, ``upper`` is exclusive.
    """
    return Integer(lower, upper).loguniform(base)


def qrandint(lower: int, upper: int, q: int = 1):
    """Sample an integer value uniformly between ``lower`` and ``upper``.
    ``lower`` is inclusive, ``upper`` is also inclusive (!).
    The value will be quantized, i.e. rounded to an integer increment of ``q``.
    Quantization makes the upper bound inclusive.
    """
    return Integer(lower, upper).uniform().quantized(q)


def qlograndint(lower: int, upper: int, q: int, base: float = 10):
    """Sample an integer value log-uniformly between ``lower`` and ``upper``,
    with ``base`` being the base of logarithm.
    ``lower`` is inclusive, ``upper`` is also inclusive (!).
    The value will be quantized, i.e. rounded to an integer increment of ``q``.
    Quantization makes the upper bound inclusive.
    """
    return Integer(lower, upper).loguniform(base).quantized(q)


def randn(mean: float = 0., sd: float = 1.):
    """Sample a float value normally with ``mean`` and ``sd``.
    Args:
        mean (float): Mean of the normal distribution. Defaults to 0.
        sd (float): SD of the normal distribution. Defaults to 1.
    """
    return Float(None, None).normal(mean, sd)


def qrandn(mean: float, sd: float, q: float):
    """Sample a float value normally with ``mean`` and ``sd``.
    The value will be quantized, i.e. rounded to an integer increment of ``q``.
    Args:
        mean (float): Mean of the normal distribution.
        sd (float): SD of the normal distribution.
        q (float): Quantization number. The result will be rounded to an
            integer increment of this value.
    """
    return Float(None, None).normal(mean, sd).quantized(q)


def polynomial_expansion_set(init_monomials: set, highest_poly_order: int = None,
                             allow_self_inter: bool = False):

    return PolynomialExpansionSet(init_monomials, highest_poly_order, allow_self_inter)



'''
Copyright 2020 The Ray Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This source file is adapted here because ray does not fully support Windows.

Copyright (c) Microsoft Corporation.
'''
import copy
import logging
from typing import Any, Dict, Optional, Union, List, Tuple
import pickle
from .variant_generator import parse_spec_vars
from ..tune.sample import Categorical, Domain, Float, Integer, LogUniform, \
    Quantized, Uniform
from ..tune.trial import flatten_dict, unflatten_dict

logger = logging.getLogger(__name__)

UNRESOLVED_SEARCH_SPACE = str(
    "You passed a `{par}` parameter to {cls} that contained unresolved search "
    "space definitions. {cls} should however be instantiated with fully "
    "configured search spaces only. To use Ray Tune's automatic search space "
    "conversion, pass the space definition as part of the `config` argument "
    "to `tune.run()` instead.")

UNDEFINED_SEARCH_SPACE = str(
    "Trying to sample a configuration from {cls}, but no search "
    "space has been defined. Either pass the `{space}` argument when "
    "instantiating the search algorithm, or pass a `config` to "
    "`tune.run()`.")

UNDEFINED_METRIC_MODE = str(
    "Trying to sample a configuration from {cls}, but the `metric` "
    "({metric}) or `mode` ({mode}) parameters have not been set. "
    "Either pass these arguments when instantiating the search algorithm, "
    "or pass them to `tune.run()`.")


class Searcher:
    """Abstract class for wrapping suggesting algorithms.
    Custom algorithms can extend this class easily by overriding the
    `suggest` method provide generated parameters for the trials.
    Any subclass that implements ``__init__`` must also call the
    constructor of this class: ``super(Subclass, self).__init__(...)``.
    To track suggestions and their corresponding evaluations, the method
    `suggest` will be passed a trial_id, which will be used in
    subsequent notifications.
    Not all implementations support multi objectives.
    Args:
        metric (str or list): The training result objective value attribute. If
            list then list of training result objective value attributes
        mode (str or list): If string One of {min, max}. If list then
            list of max and min, determines whether objective is minimizing
            or maximizing the metric attribute. Must match type of metric.
    .. code-block:: python
        class ExampleSearch(Searcher):
            def __init__(self, metric="mean_loss", mode="min", **kwargs):
                super(ExampleSearch, self).__init__(
                    metric=metric, mode=mode, **kwargs)
                self.optimizer = Optimizer()
                self.configurations = {}
            def suggest(self, trial_id):
                configuration = self.optimizer.query()
                self.configurations[trial_id] = configuration
            def on_trial_complete(self, trial_id, result, **kwargs):
                configuration = self.configurations[trial_id]
                if result and self.metric in result:
                    self.optimizer.update(configuration, result[self.metric])
        tune.run(trainable_function, search_alg=ExampleSearch())
    """
    FINISHED = "FINISHED"
    CKPT_FILE_TMPL = "searcher-state-{}.pkl"

    def __init__(self,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 max_concurrent: Optional[int] = None,
                 use_early_stopped_trials: Optional[bool] = None):
        if use_early_stopped_trials is False:
            raise DeprecationWarning(
                "Early stopped trials are now always used. If this is a "
                "problem, file an issue: https://github.com/ray-project/ray.")
        if max_concurrent is not None:
            logger.warning(
                "DeprecationWarning: `max_concurrent` is deprecated for this "
                "search algorithm. Use tune.suggest.ConcurrencyLimiter() "
                "instead. This will raise an error in future versions of Ray.")

        self._metric = metric
        self._mode = mode

        if not mode or not metric:
            # Early return to avoid assertions
            return

        assert isinstance(
            metric, type(mode)), "metric and mode must be of the same type"
        if isinstance(mode, str):
            assert mode in ["min", "max"
                            ], "if `mode` is a str must be 'min' or 'max'!"
        elif isinstance(mode, list):
            assert len(mode) == len(
                metric), "Metric and mode must be the same length"
            assert all(mod in ["min", "max", "obs"] for mod in
                       mode), "All of mode must be 'min' or 'max' or 'obs'!"
        else:
            raise ValueError("Mode most either be a list or string")

    def set_search_properties(self, metric: Optional[str], mode: Optional[str],
                              config: Dict) -> bool:
        """Pass search properties to searcher.
        This method acts as an alternative to instantiating search algorithms
        with their own specific search spaces. Instead they can accept a
        Tune config through this method. A searcher should return ``True``
        if setting the config was successful, or ``False`` if it was
        unsuccessful, e.g. when the search space has already been set.
        Args:
            metric (str): Metric to optimize
            mode (str): One of ["min", "max"]. Direction to optimize.
            config (dict): Tune config dict.
        """
        return False

    def on_trial_result(self, trial_id: str, result: Dict):
        """Optional notification for result during training.
        Note that by default, the result dict may include NaNs or
        may not include the optimization metric. It is up to the
        subclass implementation to preprocess the result to
        avoid breaking the optimization process.
        Args:
            trial_id (str): A unique string ID for the trial.
            result (dict): Dictionary of metrics for current training progress.
                Note that the result dict may include NaNs or
                may not include the optimization metric. It is up to the
                subclass implementation to preprocess the result to
                avoid breaking the optimization process.
        """
        pass

    def on_trial_complete(self,
                          trial_id: str,
                          result: Optional[Dict] = None,
                          error: bool = False):
        """Notification for the completion of trial.
        Typically, this method is used for notifying the underlying
        optimizer of the result.
        Args:
            trial_id (str): A unique string ID for the trial.
            result (dict): Dictionary of metrics for current training progress.
                Note that the result dict may include NaNs or
                may not include the optimization metric. It is up to the
                subclass implementation to preprocess the result to
                avoid breaking the optimization process. Upon errors, this
                may also be None.
            error (bool): True if the training process raised an error.
        """
        raise NotImplementedError

    def suggest(self, trial_id: str) -> Optional[Dict]:
        """Queries the algorithm to retrieve the next set of parameters.
        Arguments:
            trial_id (str): Trial ID used for subsequent notifications.
        Returns:
            dict | FINISHED | None: Configuration for a trial, if possible.
                If FINISHED is returned, Tune will be notified that
                no more suggestions/configurations will be provided.
                If None is returned, Tune will skip the querying of the
                searcher for this step.
        """
        raise NotImplementedError

    def save(self, checkpoint_path: str):
        """Save state to path for this search algorithm.
        Args:
            checkpoint_path (str): File where the search algorithm
                state is saved. This path should be used later when
                restoring from file.
        Example:
        .. code-block:: python
            search_alg = Searcher(...)
            analysis = tune.run(
                cost,
                num_samples=5,
                search_alg=search_alg,
                name=self.experiment_name,
                local_dir=self.tmpdir)
            search_alg.save("./my_favorite_path.pkl")
        .. versionchanged:: 0.8.7
            Save is automatically called by `tune.run`. You can use
            `restore_from_dir` to restore from an experiment directory
            such as `~/ray_results/trainable`.
        """
        raise NotImplementedError

    def restore(self, checkpoint_path: str):
        """Restore state for this search algorithm
        Args:
            checkpoint_path (str): File where the search algorithm
                state is saved. This path should be the same
                as the one provided to "save".
        Example:
        .. code-block:: python
            search_alg.save("./my_favorite_path.pkl")
            search_alg2 = Searcher(...)
            search_alg2 = ConcurrencyLimiter(search_alg2, 1)
            search_alg2.restore(checkpoint_path)
            tune.run(cost, num_samples=5, search_alg=search_alg2)
        """
        raise NotImplementedError

    def get_state(self) -> Dict:
        raise NotImplementedError

    def set_state(self, state: Dict):
        raise NotImplementedError

    @property
    def metric(self) -> str:
        """The training result objective value attribute."""
        return self._metric

    @property
    def mode(self) -> str:
        """Specifies if minimizing or maximizing the metric."""
        return self._mode


class ConcurrencyLimiter(Searcher):
    """A wrapper algorithm for limiting the number of concurrent trials.
    Args:
        searcher (Searcher): Searcher object that the
            ConcurrencyLimiter will manage.
        max_concurrent (int): Maximum concurrent samples from the underlying
            searcher.
        batch (bool): Whether to wait for all concurrent samples
            to finish before updating the underlying searcher.
    Example:
    .. code-block:: python
        from ray.tune.suggest import ConcurrencyLimiter
        search_alg = HyperOptSearch(metric="accuracy")
        search_alg = ConcurrencyLimiter(search_alg, max_concurrent=2)
        tune.run(trainable, search_alg=search_alg)
    """

    def __init__(self,
                 searcher: Searcher,
                 max_concurrent: int,
                 batch: bool = False):
        assert type(max_concurrent) is int and max_concurrent > 0
        self.searcher = searcher
        self.max_concurrent = max_concurrent
        self.batch = batch
        self.live_trials = set()
        self.cached_results = {}
        super(ConcurrencyLimiter, self).__init__(
            metric=self.searcher.metric, mode=self.searcher.mode)

    def suggest(self, trial_id: str) -> Optional[Dict]:
        assert trial_id not in self.live_trials, (
            f"Trial ID {trial_id} must be unique: already found in set.")
        if len(self.live_trials) >= self.max_concurrent:
            logger.debug(
                f"Not providing a suggestion for {trial_id} due to "
                "concurrency limit: %s/%s.", len(self.live_trials),
                self.max_concurrent)
            return

        suggestion = self.searcher.suggest(trial_id)
        if suggestion not in (None, Searcher.FINISHED):
            self.live_trials.add(trial_id)
        return suggestion

    def on_trial_complete(self,
                          trial_id: str,
                          result: Optional[Dict] = None,
                          error: bool = False):
        if trial_id not in self.live_trials:
            return
        elif self.batch:
            self.cached_results[trial_id] = (result, error)
            if len(self.cached_results) == self.max_concurrent:
                # Update the underlying searcher once the
                # full batch is completed.
                for trial_id, (result, error) in self.cached_results.items():
                    self.searcher.on_trial_complete(
                        trial_id, result=result, error=error)
                    self.live_trials.remove(trial_id)
                self.cached_results = {}
            else:
                return
        else:
            self.searcher.on_trial_complete(
                trial_id, result=result, error=error)
            self.live_trials.remove(trial_id)

    def get_state(self) -> Dict:
        state = self.__dict__.copy()
        del state["searcher"]
        return copy.deepcopy(state)

    def set_state(self, state: Dict):
        self.__dict__.update(state)

    def save(self, checkpoint_path: str):
        self.searcher.save(checkpoint_path)

    def restore(self, checkpoint_path: str):
        self.searcher.restore(checkpoint_path)

    def on_pause(self, trial_id: str):
        self.searcher.on_pause(trial_id)

    def on_unpause(self, trial_id: str):
        self.searcher.on_unpause(trial_id)

    def set_search_properties(self, metric: Optional[str], mode: Optional[str],
                              config: Dict) -> bool:
        return self.searcher.set_search_properties(metric, mode, config)


try:
    import optuna as ot
    from optuna.trial import TrialState as OptunaTrialState
    from optuna.samplers import BaseSampler
except ImportError:
    ot = None
    OptunaTrialState = None
    BaseSampler = None


# (Optional) Default (anonymous) metric when using tune.report(x)
DEFAULT_METRIC = "_metric"

# (Auto-filled) The index of this training iteration.
TRAINING_ITERATION = "training_iteration"


class OptunaSearch(Searcher):
    """A wrapper around Optuna to provide trial suggestions.
    `Optuna <https://optuna.org/>`_ is a hyperparameter optimization library.
    In contrast to other libraries, it employs define-by-run style
    hyperparameter definitions.
    This Searcher is a thin wrapper around Optuna's search algorithms.
    You can pass any Optuna sampler, which will be used to generate
    hyperparameter suggestions.
    Please note that this wrapper does not support define-by-run, so the
    search space will be configured before running the optimization. You will
    also need to use a Tune trainable (e.g. using the function API) with
    this wrapper.
    For defining the search space, use ``ray.tune.suggest.optuna.param``
    (see example).
    Args:
        space (list): Hyperparameter search space definition for Optuna's
            sampler. This is a list, and samples for the parameters will
            be obtained in order.
        metric (str): The training result objective value attribute. If None
            but a mode was passed, the anonymous metric `_metric` will be used
            per default.
        mode (str): One of {min, max}. Determines whether objective is
            minimizing or maximizing the metric attribute.
        points_to_evaluate (list): Initial parameter suggestions to be run
            first. This is for when you already have some good parameters
            you want to run first to help the algorithm make better suggestions
            for future parameters. Needs to be a list of dicts containing the
            configurations.
        sampler (optuna.samplers.BaseSampler): Optuna sampler used to
            draw hyperparameter configurations. Defaults to ``TPESampler``.
        seed (int): Seed to initialize sampler with. This parameter is only
            used when ``sampler=None``. In all other cases, the sampler
            you pass should be initialized with the seed already.
        evaluated_rewards (list): If you have previously evaluated the
            parameters passed in as points_to_evaluate you can avoid
            re-running those trials by passing in the reward attributes
            as a list so the optimiser can be told the results without
            needing to re-compute the trial. Must be the same length as
            points_to_evaluate.
    Tune automatically converts search spaces to Optuna's format:
    .. code-block:: python
        from ray.tune.suggest.optuna import OptunaSearch
        config = {
            "a": tune.uniform(6, 8)
            "b": tune.loguniform(1e-4, 1e-2)
        }
        optuna_search = OptunaSearch(
            metric="loss",
            mode="min")
        tune.run(trainable, config=config, search_alg=optuna_search)
    If you would like to pass the search space manually, the code would
    look like this:
    .. code-block:: python
        from ray.tune.suggest.optuna import OptunaSearch
        import optuna
        config = {
            "a": optuna.distributions.UniformDistribution(6, 8),
            "b": optuna.distributions.LogUniformDistribution(1e-4, 1e-2),
        }
        optuna_search = OptunaSearch(
            space,
            metric="loss",
            mode="min")
        tune.run(trainable, search_alg=optuna_search)
    .. versionadded:: 0.8.8
    """

    def __init__(self,
                 space: Optional[Union[Dict, List[Tuple]]] = None,
                 metric: Optional[str] = None,
                 mode: Optional[str] = None,
                 points_to_evaluate: Optional[List[Dict]] = None,
                 sampler: Optional[BaseSampler] = None,
                 seed: Optional[int] = None,
                 evaluated_rewards: Optional[List] = None):
        assert ot is not None, (
            "Optuna must be installed! Run `pip install optuna`.")
        super(OptunaSearch, self).__init__(
            metric=metric,
            mode=mode,
            max_concurrent=None,
            use_early_stopped_trials=None)

        if isinstance(space, dict) and space:
            resolved_vars, domain_vars, grid_vars = parse_spec_vars(space)
            if domain_vars or grid_vars:
                logger.warning(
                    UNRESOLVED_SEARCH_SPACE.format(
                        par="space", cls=type(self).__name__))
                space = self.convert_search_space(space)
            else:
                # Flatten to support nested dicts
                space = flatten_dict(space, "/")

        # Deprecate: 1.5
        if isinstance(space, list):
            logger.warning(
                "Passing lists of `param.suggest_*()` calls to OptunaSearch "
                "as a search space is deprecated and will be removed in "
                "a future release of Ray. Please pass a dict mapping "
                "to `optuna.distributions` objects instead.")

        self._space = space

        self._points_to_evaluate = points_to_evaluate or []
        self._evaluated_rewards = evaluated_rewards

        self._study_name = "optuna"  # Fixed study name for in-memory storage

        if sampler and seed:
            logger.warning(
                "You passed an initialized sampler to `OptunaSearch`. The "
                "`seed` parameter has to be passed to the sampler directly "
                "and will be ignored.")

        self._sampler = sampler or ot.samplers.TPESampler(seed=seed)

        assert isinstance(self._sampler, BaseSampler), \
            "You can only pass an instance of `optuna.samplers.BaseSampler` " \
            "as a sampler to `OptunaSearcher`."

        self._ot_trials = {}
        self._ot_study = None
        if self._space:
            self._setup_study(mode)

    def _setup_study(self, mode: str):
        if self._metric is None and self._mode:
            # If only a mode was passed, use anonymous metric
            self._metric = DEFAULT_METRIC

        pruner = ot.pruners.NopPruner()
        storage = ot.storages.InMemoryStorage()

        self._ot_study = ot.study.create_study(
            storage=storage,
            sampler=self._sampler,
            pruner=pruner,
            study_name=self._study_name,
            direction="minimize" if mode == "min" else "maximize",
            load_if_exists=True)

        if self._points_to_evaluate:
            if self._evaluated_rewards:
                for point, reward in zip(self._points_to_evaluate,
                                         self._evaluated_rewards):
                    self.add_evaluated_point(point, reward)
            else:
                for point in self._points_to_evaluate:
                    self._ot_study.enqueue_trial(point)

    def set_search_properties(self, metric: Optional[str], mode: Optional[str],
                              config: Dict) -> bool:
        if self._space:
            return False
        space = self.convert_search_space(config)
        self._space = space
        if metric:
            self._metric = metric
        if mode:
            self._mode = mode

        self._setup_study(mode)
        return True

    def suggest(self, trial_id: str) -> Optional[Dict]:
        if not self._space:
            raise RuntimeError(
                UNDEFINED_SEARCH_SPACE.format(
                    cls=self.__class__.__name__, space="space"))
        if not self._metric or not self._mode:
            raise RuntimeError(
                UNDEFINED_METRIC_MODE.format(
                    cls=self.__class__.__name__,
                    metric=self._metric,
                    mode=self._mode))

        if isinstance(self._space, list):
            # Keep for backwards compatibility
            # Deprecate: 1.5
            if trial_id not in self._ot_trials:
                self._ot_trials[trial_id] = self._ot_study.ask()

            ot_trial = self._ot_trials[trial_id]

            # getattr will fetch the trial.suggest_ function on Optuna trials
            params = {
                args[0] if len(args) > 0 else kwargs["name"]: getattr(
                    ot_trial, fn)(*args, **kwargs)
                for (fn, args, kwargs) in self._space
            }
        else:
            # Use Optuna ask interface (since version 2.6.0)
            if trial_id not in self._ot_trials:
                self._ot_trials[trial_id] = self._ot_study.ask(
                    fixed_distributions=self._space)
            ot_trial = self._ot_trials[trial_id]
            params = ot_trial.params

        return unflatten_dict(params)

    def on_trial_result(self, trial_id: str, result: Dict):
        metric = result[self.metric]
        step = result[TRAINING_ITERATION]
        ot_trial = self._ot_trials[trial_id]
        ot_trial.report(metric, step)

    def on_trial_complete(self,
                          trial_id: str,
                          result: Optional[Dict] = None,
                          error: bool = False):
        ot_trial = self._ot_trials[trial_id]

        val = result.get(self.metric, None) if result else None
        ot_trial_state = OptunaTrialState.COMPLETE
        if val is None:
            if error:
                ot_trial_state = OptunaTrialState.FAIL
            else:
                ot_trial_state = OptunaTrialState.PRUNED
        try:
            self._ot_study.tell(ot_trial, val, state=ot_trial_state)
        except ValueError as exc:
            logger.warning(exc)  # E.g. if NaN was reported

    def add_evaluated_point(self,
                            parameters: Dict,
                            value: float,
                            error: bool = False,
                            pruned: bool = False,
                            intermediate_values: Optional[List[float]] = None):
        if not self._space:
            raise RuntimeError(
                UNDEFINED_SEARCH_SPACE.format(
                    cls=self.__class__.__name__, space="space"))
        if not self._metric or not self._mode:
            raise RuntimeError(
                UNDEFINED_METRIC_MODE.format(
                    cls=self.__class__.__name__,
                    metric=self._metric,
                    mode=self._mode))

        ot_trial_state = OptunaTrialState.COMPLETE
        if error:
            ot_trial_state = OptunaTrialState.FAIL
        elif pruned:
            ot_trial_state = OptunaTrialState.PRUNED

        if intermediate_values:
            intermediate_values_dict = {
                i: value
                for i, value in enumerate(intermediate_values)
            }
        else:
            intermediate_values_dict = None

        trial = ot.trial.create_trial(
            state=ot_trial_state,
            value=value,
            params=parameters,
            distributions=self._space,
            intermediate_values=intermediate_values_dict)

        self._ot_study.add_trial(trial)

    def save(self, checkpoint_path: str):
        save_object = (self._sampler, self._ot_trials, self._ot_study,
                       self._points_to_evaluate, self._evaluated_rewards)
        with open(checkpoint_path, "wb") as outputFile:
            pickle.dump(save_object, outputFile)

    def restore(self, checkpoint_path: str):
        with open(checkpoint_path, "rb") as inputFile:
            save_object = pickle.load(inputFile)
        if len(save_object) == 5:
            self._sampler, self._ot_trials, self._ot_study, \
                self._points_to_evaluate, self._evaluated_rewards = save_object
        else:
            # Backwards compatibility
            self._sampler, self._ot_trials, self._ot_study, \
                self._points_to_evaluate = save_object

    @staticmethod
    def convert_search_space(spec: Dict) -> Dict[str, Any]:
        resolved_vars, domain_vars, grid_vars = parse_spec_vars(spec)

        if not domain_vars and not grid_vars:
            return {}

        if grid_vars:
            raise ValueError(
                "Grid search parameters cannot be automatically converted "
                "to an Optuna search space.")

        # Flatten and resolve again after checking for grid search.
        spec = flatten_dict(spec, prevent_delimiter=True)
        resolved_vars, domain_vars, grid_vars = parse_spec_vars(spec)

        def resolve_value(domain: Domain) -> ot.distributions.BaseDistribution:
            quantize = None

            sampler = domain.get_sampler()
            if isinstance(sampler, Quantized):
                quantize = sampler.q
                sampler = sampler.sampler
                if isinstance(sampler, LogUniform):
                    logger.warning(
                        "Optuna does not handle quantization in loguniform "
                        "sampling. The parameter will be passed but it will "
                        "probably be ignored.")

            if isinstance(domain, Float):
                if isinstance(sampler, LogUniform):
                    if quantize:
                        logger.warning(
                            "Optuna does not support both quantization and "
                            "sampling from LogUniform. Dropped quantization.")
                    return ot.distributions.LogUniformDistribution(
                        domain.lower, domain.upper)

                elif isinstance(sampler, Uniform):
                    if quantize:
                        return ot.distributions.DiscreteUniformDistribution(
                            domain.lower, domain.upper, quantize)
                    return ot.distributions.UniformDistribution(
                        domain.lower, domain.upper)

            elif isinstance(domain, Integer):
                if isinstance(sampler, LogUniform):
                    return ot.distributions.IntLogUniformDistribution(
                        domain.lower, domain.upper - 1, step=quantize or 1)
                elif isinstance(sampler, Uniform):
                    # Upper bound should be inclusive for quantization and
                    # exclusive otherwise
                    return ot.distributions.IntUniformDistribution(
                        domain.lower,
                        domain.upper - int(bool(not quantize)),
                        step=quantize or 1)
            elif isinstance(domain, Categorical):
                if isinstance(sampler, Uniform):
                    return ot.distributions.CategoricalDistribution(
                        domain.categories)

            raise ValueError(
                "Optuna search does not support parameters of type "
                "`{}` with samplers of type `{}`".format(
                    type(domain).__name__,
                    type(domain.sampler).__name__))

        # Parameter name is e.g. "a/b/c" for nested dicts
        values = {
            "/".join(path): resolve_value(domain)
            for path, domain in domain_vars
        }

        return values


import os
import subprocess
import hashlib
from time import time
import json


class WandbUtils:

    # Documentation on the wandb setting:
    # There are two ways to initialize wandb in tune.run:
    # (1) using WandbLoggerCallback, by adding the following argument to tune.run:
    #     callbacks=[WandbLoggerCallback(
    #                  project="hpo",
    #                  api_key = os.environ["WANDB_API_KEY"],
    #                  group = os.environ["WANDB_RUN_GROUP"],
    #                  log_config=True)]
    # (2) using wandb_mixin decorator (the current implementation)
    # The current implementation uses (2) because (1) has the following bug.
    # In Ray 1.2, when using WandbLoggerCallback + setting time limit using the time_budget_s argument,
    # A bug exists which is the previous run will not clear the cache after tune.run returns. After the
    # later run has already starts, some zombie trials in the previous run remain in the memory and never stop.
    # This bug can be reproduced by switching to (1) by adding the above callbacks argument
    # and removing the wandb_mixin decorator
    # https://docs.ray.io/en/master/tune/tutorials/tune-wandb.html

    def __init__(self,
                 is_wandb_on=False,
                 wandb_key_path=None,
                 jobid_config=None):
        if is_wandb_on:
            wandb_key = WandbUtils.get_wandb_key(wandb_key_path)
            if wandb_key != "":
                subprocess.run(["wandb", "login", "--relogin", wandb_key])
            os.environ["WANDB_API_KEY"] = wandb_key
            os.environ["WANDB_MODE"] = "online"
        else:
            os.environ["WANDB_MODE"] = "disabled"
        self.jobid_config = jobid_config

    @staticmethod
    def get_wandb_key(key_path):
        try:
            try:
                key_json = json.load(open(os.path.join(key_path, "key.json"), "r"))
                wandb_key = key_json["wandb_key"]
                return wandb_key
            except FileNotFoundError:
                print("Cannot use wandb module because key.json is not found under key_path")
                return ""
        except KeyError:
            print("Cannot use wandb module because wandb key is not specified")
            return ""

    def set_wandb_per_trial(self):
        print("before wandb.init\n\n\n")
        try:
            import wandb
            try:
                if os.environ["WANDB_MODE"] == "online":
                    os.environ["WANDB_SILENT"] = "false"
                    return wandb.init(project=self.jobid_config.get_jobid_full_data_name(),
                                      group=self.wandb_group_name,
                                      name=str(WandbUtils._get_next_trial_ids()),
                                      settings=wandb.Settings(
                                          _disable_stats=True),
                                      reinit=False)
                else:
                    return None
            except wandb.errors.UsageError as err:
                print(err)
                return None
        except ImportError:
            print("Cannot use wandb module because wandb is not installed, run pip install wandb==0.10.26")

    @staticmethod
    def _get_next_trial_ids():
        hash = hashlib.sha1()
        hash.update(str(time()).encode('utf-8'))
        return "trial_" + hash.hexdigest()[:3]

    def set_wandb_per_run(self):
        try:
            import wandb
            os.environ["WANDB_RUN_GROUP"] = self.jobid_config.to_wandb_string() + wandb.util.generate_id()
            self.wandb_group_name = os.environ["WANDB_RUN_GROUP"]
            try:
                if os.environ["WANDB_MODE"] == "online":
                    os.environ["WANDB_SILENT"] = "false"
                    return wandb.init(project=self.jobid_config.get_jobid_full_data_name(),
                                      group=os.environ["WANDB_RUN_GROUP"],
                                      settings=wandb.Settings(
                                          _disable_stats=True),
                                      reinit=False)
                else:
                    return None
            except wandb.errors.UsageError as err:
                print(err)
                return None
        except ImportError:
            print("Cannot use wandb module because wandb is not installed, run pip install wandb==0.10.26")



import unittest

from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
import numpy as np
from flaml.automl import AutoML
from flaml.model import XGBoostSklearnEstimator
from flaml import tune


dataset = "credit-g"


class XGBoost2D(XGBoostSklearnEstimator):

    @classmethod
    def search_space(cls, data_size, task):
        upper = min(32768, int(data_size))
        return {
            'n_estimators': {
                'domain': tune.lograndint(lower=4, upper=upper),
                'init_value': 4,
            },
            'max_leaves': {
                'domain': tune.lograndint(lower=4, upper=upper),
                'init_value': 4,
            },
        }


def _test_simple(method=None, size_ratio=1.0):
    automl = AutoML()
    automl.add_learner(learner_name='XGBoost2D',
                       learner_class=XGBoost2D)

    X, y = fetch_openml(name=dataset, return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.33,
                                                        random_state=42)

    final_size = int(len(y_train) * size_ratio)
    X_train = X_train[:final_size]
    y_train = y_train[:final_size]
    automl_settings = {
        "estimator_list": ['XGBoost2D'],
        # "metric": 'accuracy',
        "task": 'classification',
        "log_file_name": f"test/xgboost2d_{dataset}_{method}_{final_size}.log",
        # "model_history": True,
        # "log_training_metric": True,
        # "split_type": split_type,
        "n_jobs": 1,
        "hpo_method": method,
        "log_type": "all",
        "time_budget": 3600,
    }
    automl.fit(X_train=X_train, y_train=y_train, **automl_settings)


def _test_grid_1():
    _test_simple(method="grid", size_ratio=1.0 / 3.0)


def _test_grid_2():
    _test_simple(method="grid", size_ratio=2.0 / 3.0)


def _test_grid_4():
    _test_simple(method="grid", size_ratio=0.5)


def _test_grid_3():
    _test_simple(method="grid", size_ratio=1.0)


if __name__ == "__main__":
    unittest.main()



import os
import shutil
import tempfile
import unittest
import numpy as np
from flaml.searcher.suggestion import ConcurrencyLimiter
from flaml import tune
from flaml import CFO
from flaml import BlendSearch


class AbstractWarmStartTest:
    def setUp(self):
        # ray.init(num_cpus=1, local_mode=True)
        self.tmpdir = tempfile.mkdtemp()
        self.experiment_name = "searcher-state-Test.pkl"

    def tearDown(self):
        shutil.rmtree(self.tmpdir)
        # ray.shutdown()

    def set_basic_conf(self):
        raise NotImplementedError()

    def run_part_from_scratch(self):
        np.random.seed(162)
        search_alg, cost = self.set_basic_conf()
        search_alg = ConcurrencyLimiter(search_alg, 1)
        results_exp_1 = tune.run(
            cost,
            num_samples=5,
            search_alg=search_alg,
            verbose=0,
            local_dir=self.tmpdir)
        checkpoint_path = os.path.join(self.tmpdir, self.experiment_name)
        search_alg.save(checkpoint_path)
        return results_exp_1, np.random.get_state(), checkpoint_path

    def run_explicit_restore(self, random_state, checkpoint_path):
        np.random.set_state(random_state)
        search_alg2, cost = self.set_basic_conf()
        search_alg2 = ConcurrencyLimiter(search_alg2, 1)
        search_alg2.restore(checkpoint_path)
        return tune.run(cost, num_samples=5, search_alg=search_alg2, verbose=0)

    def run_full(self):
        np.random.seed(162)
        search_alg3, cost = self.set_basic_conf()
        search_alg3 = ConcurrencyLimiter(search_alg3, 1)
        return tune.run(
            cost, num_samples=10, search_alg=search_alg3, verbose=0)

    def testReproduce(self):
        results_exp_1, _, _ = self.run_part_from_scratch()
        results_exp_2, _, _ = self.run_part_from_scratch()
        trials_1_config = [trial.config for trial in results_exp_1.trials]
        trials_2_config = [trial.config for trial in results_exp_2.trials]
        self.assertEqual(trials_1_config, trials_2_config)

    def testWarmStart(self):
        results_exp_1, r_state, checkpoint_path = self.run_part_from_scratch()
        results_exp_2 = self.run_explicit_restore(r_state, checkpoint_path)
        results_exp_3 = self.run_full()
        trials_1_config = [trial.config for trial in results_exp_1.trials]
        trials_2_config = [trial.config for trial in results_exp_2.trials]
        trials_3_config = [trial.config for trial in results_exp_3.trials]
        self.assertEqual(trials_1_config + trials_2_config, trials_3_config)


class CFOWarmStartTest(AbstractWarmStartTest, unittest.TestCase):
    def set_basic_conf(self):
        space = {
            "height": tune.uniform(-100, 100),
            "width": tune.randint(0, 100),
        }

        def cost(param):
            tune.report(loss=(param["height"] - 14)**2 - abs(param["width"] - 3))

        search_alg = CFO(
            space=space,
            metric="loss",
            mode="min",
            seed=20,
        )

        return search_alg, cost

# # # Not doing test for BS because of problems with random seed in OptunaSearch
# class BlendsearchWarmStartTest(AbstractWarmStartTest, unittest.TestCase):
#     def set_basic_conf(self):
#         space = {
#             "height": tune.uniform(-100, 100),
#             "width": tune.randint(0, 100),
#         }

#         def cost(param):
#             tune.report(loss=(param["height"] - 14)**2 - abs(param["width"] - 3))

#         search_alg = BlendSearch(
#             space=space,
#             metric="loss",
#             mode="min",
#             seed=20,
#         )

#         return search_alg, cost
